from __future__ import annotations

from typing import Union

import pwndbg.aglib
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod
import pwndbg.lib.cache
import pwndbg.lib.memory
from pwndbg.dbg_mod import EventType
from pwndbg.dbg_mod import TypeCode
from pwndbg.lib import TypeNotFoundError
from pwndbg.lib.memory import PAGE_SIZE

GdbDict = dict[str, Union["GdbDict", int]]


MMAP_MIN_ADDR = 0x8000


def read(addr: int, count: int, partial: bool = False) -> bytearray:
    """read(addr, count, partial=False) -> bytearray

    Read memory from the program being debugged.

    Arguments:
        addr: Address to read
        count: Number of bytes to read
        partial: Whether less than ``count`` bytes can be returned

    Returns:
        `bytearray` The memory at the specified address,
        or ``None``.
    """
    return pwndbg.dbg.selected_inferior().read_memory(address=addr, size=count, partial=partial)


def readtype(type: pwndbg.dbg_mod.Type, addr: int) -> int:
    """readtype(type, addr) -> int

    Reads an integer-type (e.g. ``uint64``) and returns a Python
    native integer representation of the same.

    Arguments:
        type: GDB type to read
        addr: Address at which the value to be read resides

    Raises:
        TypeNotFoundError: If the type does not exist in the debugger.

    Returns:
        `int`
    """
    return int(get_typed_pointer_value(type, addr))


def write(addr: int, data: str | bytes | bytearray) -> None:
    """write(addr, data)

    Writes data into the memory of the process being debugged.

    Arguments:
        addr: Address to write
        data: Data to write
    """
    if isinstance(data, str):
        data = bytes(data, "utf8")

    pwndbg.dbg.selected_inferior().write_memory(address=addr, data=bytearray(data), partial=False)


def peek(address: int) -> bytearray | None:
    """peek(address) -> bytearray

    Read one byte from the specified address.

    Arguments:
        address: Address to read

    Returns:
        `bytearray` A single byte of data, or ``None`` if the
        address cannot be read.
    """
    try:
        return read(address, 1)
    except Exception:
        pass
    return None


@pwndbg.lib.cache.cache_until("stop")
def is_readable_address(address: int) -> bool:
    """is_readable_address(address) -> bool

    Check if the address can be read by GDB.

    Arguments:
        address: Address to read

    Returns:
        `bool`: Whether the address is readable.
    """
    # We use vmmap to check before `peek()` because accessing memory for embedded targets might be slow and expensive.
    return pwndbg.aglib.vmmap.find(address) is not None and peek(address) is not None


def poke(address: int) -> bool:
    """poke(address)

    Checks whether an address is writable.

    Arguments:
        address: Address to check

    Returns:
        `bool`: Whether the address is writable.
    """
    c = peek(address)
    if c is None:
        return False
    try:
        # Suspending mem_changed event during poke speeds up things when vmmaps are explored
        # (e.g. when stepping through remote processes run with QEMU)
        # The suspension prevents the clearing of the disasm instruction cache
        # by `aglib.disasm.clear_on_reg_mem_change`
        pwndbg.dbg.suspend_events(EventType.MEMORY_CHANGED)
        write(address, c)
    except Exception:
        return False
    finally:
        pwndbg.dbg.resume_events(EventType.MEMORY_CHANGED)

    return True


def string(addr: int, max: int = 4096) -> bytearray:
    """Reads a null-terminated string from memory.

    Arguments:
        addr: Address to read from
        max: Maximum string length (default 4096)

    Returns:
        An empty bytearray, or a NULL-terminated bytearray.
    """
    if peek(addr):
        data = read(addr, max, partial=True)

        try:
            return data[: data.index(b"\x00")]
        except ValueError:
            pass

    return bytearray()


def byte(addr: int) -> int:
    """byte(addr) -> int

    Read one byte at the specified address
    """
    return readtype(pwndbg.aglib.typeinfo.uchar, addr)


def uchar(addr: int) -> int:
    """uchar(addr) -> int

    Read one ``unsigned char`` at the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.uchar, addr)


def ushort(addr: int) -> int:
    """ushort(addr) -> int

    Read one ``unisgned short`` at the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.ushort, addr)


def uint(addr: int) -> int:
    """uint(addr) -> int

    Read one ``unsigned int`` at the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.uint, addr)


def read_pointer_width(addr: int) -> int:
    """
    Read one pointer-width integer at the specified address.

    Raises:
        pwndbg.dbg_mod.Error: if memory read fails.
    """
    return pwndbg.aglib.arch.unpack(read(addr, pwndbg.aglib.arch.ptrsize))


def u8(addr: int) -> int:
    """u8(addr) -> int

    Read one ``uint8_t`` from the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.uint8, addr)


def u16(addr: int) -> int:
    """u16(addr) -> int

    Read one ``uint16_t`` from the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.uint16, addr)


def u32(addr: int) -> int:
    """u32(addr) -> int

    Read one ``uint32_t`` from the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.uint32, addr)


def u64(addr: int) -> int:
    """u64(addr) -> int

    Read one ``uint64_t`` from the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.uint64, addr)


def u(addr: int, size: int | None = None) -> int:
    """u(addr, size=None) -> int

    Read one ``unsigned`` integer from the specified address,
    with the bit-width specified by ``size``, which defaults
    to the pointer width.
    """
    if size is None:
        size = pwndbg.aglib.arch.ptrbits
    return {8: u8, 16: u16, 32: u32, 64: u64}[size](addr)


def s8(addr: int) -> int:
    """s8(addr) -> int

    Read one ``int8_t`` from the specified address
    """
    return readtype(pwndbg.aglib.typeinfo.int8, addr)


def s16(addr: int) -> int:
    """s16(addr) -> int

    Read one ``int16_t`` from the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.int16, addr)


def s32(addr: int) -> int:
    """s32(addr) -> int

    Read one ``int32_t`` from the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.int32, addr)


def s64(addr: int) -> int:
    """s64(addr) -> int

    Read one ``int64_t`` from the specified address.
    """
    return readtype(pwndbg.aglib.typeinfo.int64, addr)


def sint(addr: int) -> int:
    """
    Read one `signed int` from the specified
    address.
    """
    return readtype(pwndbg.aglib.typeinfo.sint, addr)


def cast_pointer(
    type: pwndbg.dbg_mod.Type, addr: int | pwndbg.dbg_mod.Value
) -> pwndbg.dbg_mod.Value:
    """Create a Value containing given address and cast it to the pointer of specified type"""
    if isinstance(addr, int):
        addr = pwndbg.dbg.selected_inferior().create_value(addr)
    return addr.cast(type.pointer())


def get_typed_pointer(
    type: str | pwndbg.dbg_mod.Type, addr: int | pwndbg.dbg_mod.Value
) -> pwndbg.dbg_mod.Value:
    """
    Look up a type by name if necessary and return a Value of addr cast to that type.

    Raises:
        TypeNotFoundError: If the type does not exist in the debugger.
    """
    if addr is None:
        return None
    if isinstance(type, str):
        real_type = pwndbg.aglib.typeinfo.load(type)
        if real_type is None:
            raise TypeNotFoundError(f"Type '{type}' not found")
    elif isinstance(type, pwndbg.dbg_mod.Type):
        real_type = type
    else:
        raise ValueError(f"Invalid type: {type}")
    return cast_pointer(real_type, addr)


def get_typed_pointer_value(
    type_name: str | pwndbg.dbg_mod.Type, addr: int | pwndbg.dbg_mod.Value
) -> pwndbg.dbg_mod.Value:
    """
    Read the pointer value of addr cast to type specified by type_name.

    Raises:
        TypeNotFoundError: If the type does not exist in the debugger.
    """
    return get_typed_pointer(type_name, addr).dereference()


@pwndbg.lib.cache.cache_until("stop")
def find_upper_boundary(addr: int, max_pages: int = 1024) -> int:
    """find_upper_boundary(addr, max_pages=1024) -> int

    Brute-force search the upper boundary of a memory mapping,
    by reading the first byte of each page, until an unmapped
    page is found.
    """
    addr = pwndbg.lib.memory.page_align(int(addr))
    try:
        for _ in range(max_pages):
            read(addr, 1)
            # import sys
            # sys.stdout.write(hex(addr) + '\n')
            addr += PAGE_SIZE

            # Sanity check in case a custom GDB server/stub
            # incorrectly returns a result from read
            # (this is most likely redundant, but its ok to keep it?)
            if addr > pwndbg.aglib.arch.ptrmask:
                return pwndbg.aglib.arch.ptrmask
    except pwndbg.dbg_mod.Error:
        pass
    return addr


@pwndbg.lib.cache.cache_until("stop")
def find_lower_boundary(addr: int, max_pages: int = 1024) -> int:
    """find_lower_boundary(addr, max_pages=1024) -> int

    Brute-force search the lower boundary of a memory mapping,
    by reading the first byte of each page, until an unmapped
    page is found.
    """
    addr = pwndbg.lib.memory.page_align(int(addr))
    try:
        for _ in range(max_pages):
            read(addr, 1)
            addr -= PAGE_SIZE

            # Sanity check (see comment in find_upper_boundary)
            if addr < 0:
                return 0

    except pwndbg.dbg_mod.Error:
        addr += PAGE_SIZE
    return addr


def update_min_addr() -> None:
    global MMAP_MIN_ADDR
    MMAP_MIN_ADDR = 0 if pwndbg.aglib.qemu.is_qemu_kernel() else 0x8000


def fetch_struct_as_dictionary(
    struct_name: str,
    struct_address: int | pwndbg.dbg_mod.Value,
    include_only_fields: set[str] | None = None,
    exclude_fields: set[str] | None = None,
) -> GdbDict:
    """
    Raises:
        TypeNotFoundError: If the type does not exist in the debugger.
    """
    fetched_struct = get_typed_pointer_value("struct " + struct_name, struct_address)
    return pack_struct_into_dictionary(fetched_struct, include_only_fields, exclude_fields)


def pack_struct_into_dictionary(
    fetched_struct: pwndbg.dbg_mod.Value,
    include_only_fields: set[str] | None = None,
    exclude_fields: set[str] | None = None,
) -> GdbDict:
    struct_as_dictionary = {}

    if exclude_fields is None:
        exclude_fields = set()

    if include_only_fields is not None:
        for field_name in include_only_fields:
            key = field_name
            value = convert_pwndbg_value_to_python_value(fetched_struct[field_name])
            struct_as_dictionary[key] = value
    else:
        for index, field in enumerate(fetched_struct.type.fields()):
            if field.name is None:
                # Flatten anonymous structs/unions
                anon_type = convert_pwndbg_value_to_python_value(fetched_struct[index])
                assert isinstance(anon_type, dict)
                struct_as_dictionary.update(anon_type)
            elif field.name not in exclude_fields:
                key = field.name
                value = convert_pwndbg_value_to_python_value(fetched_struct[index])
                struct_as_dictionary[key] = value

    return struct_as_dictionary


def convert_pwndbg_value_to_python_value(dbg_value: pwndbg.dbg_mod.Value) -> int | GdbDict:
    ty = dbg_value.type.strip_typedefs()

    if ty.code == TypeCode.POINTER or ty.code == TypeCode.INT:
        return int(dbg_value)
    if ty.code == TypeCode.STRUCT:
        return pack_struct_into_dictionary(dbg_value)

    raise NotImplementedError


def resolve_renamed_struct_field(struct_name: str, possible_field_names: set[str]) -> str:
    struct_types = pwndbg.dbg.selected_inferior().types_with_name(f"struct {struct_name}")
    if len(struct_types) == 0:
        raise pwndbg.dbg_mod.Error(f"could not find type 'struct {struct_name}'")
    struct_type = struct_types[0]

    for field_name in possible_field_names:
        if struct_type.has_field(field_name):
            return field_name

    raise ValueError(f"Field name did not match any of {possible_field_names}.")


@pwndbg.lib.cache.cache_until("start", "objfile")
def is_pagefault_supported() -> bool:
    """
    This function should be called before stray memory dereferences to protect against the following situations:

    1. On embedded systems, it's not uncommon for MMIO regions to exist where memory reads might mutate the hardware/process state.
    2. On baremetal/embedded, paging doesn't always exist, so all memory is "valid" (and often initialized to zero) - this makes every value appear to be a pointer.

    As such, we disable dereferencing by default for bare metal targets.

    See more discussion here: https://github.com/pwndbg/pwndbg/pull/385
    """

    # TODO: use a better detection method
    return pwndbg.dbg.selected_inferior().is_linux()


def is_kernel(addr: int | None):
    return addr is not None and (addr >> 63 == 1) and peek(addr) is not None



from __future__ import annotations

import argparse

import pwndbg.aglib.kernel
import pwndbg.aglib.proc
import pwndbg.color.message as message
import pwndbg.commands
from pwndbg import config
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Finds the kernel virtual base address.")

parser.add_argument("-r", "--rebase", action="store_true", help="rebase loaded symbol file")
parser.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="show more information relevant to the kbase (e.g. phys addr)",
)


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
def kbase(rebase=False, verbose=False) -> None:
    if config.kernel_vmmap == "none":
        print(message.error("kbase does not work when kernel-vmmap is set to none"))
        return

    base = pwndbg.aglib.kernel.kbase()

    if base is None:
        print(message.error("Unable to locate the kernel base"))
        return

    print(message.success(f"Found virtual text base address: {hex(base)}"))

    if verbose:
        phys = pwndbg.aglib.kernel.virt_to_phys(base)
        print(message.success(f"corresponding physical address: {hex(phys)}"))

    if not rebase:
        return

    symbol_file = pwndbg.aglib.proc.exe()

    if symbol_file:
        pwndbg.dbg.selected_inferior().add_symbol_file(symbol_file, base)
    else:
        print(message.error("No symbol file is currently loaded"))



"""
The error handling logic.

Contains error and exception definitions.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


class ErrorCode(Enum):
    """
    Describes which error occured.

    Feel free to add a new one.
    """

    SUCCESS = 0
    FAILURE = 1
    """Generic failure"""
    NO_STRUCTURE_FILE = 2
    """We tried to load a structure, but the associated structure file doesn't exist."""
    NO_IMPORT = 3
    """We tried to import a python library but it is not installed."""


@dataclass
class Status:
    """
    An error handling object.
    """

    # Inspired by the LLVM Status object:
    # https://github.com/llvm/llvm-project/blob/5a4754d2ced8792be2c21eedfde1ed826f7ef64a/lldb/include/lldb/Utility/Status.h#L118

    code: ErrorCode
    """An error code associated with this error."""
    message: str = ""
    """If is_failure() returns True, this attribute contains the error message."""

    def __init__(self, code: ErrorCode = ErrorCode.SUCCESS, message: str = "") -> None:
        """
        Construct a status object.

        Passing no arguments (i.e. doing `Status()`) represents a success.
        """
        self.code = code
        self.message = message

    def is_success(self) -> bool:
        """Returns True if the Status object denotes a success."""
        return self.code == ErrorCode.SUCCESS

    def is_failure(self) -> bool:
        """Returns True if the Status object denotes a failure."""
        return self.code != ErrorCode.SUCCESS

    @staticmethod
    def succeed() -> Status:
        """Create a Status object that represents a success."""
        return Status()

    @staticmethod
    def fail(message: str = "") -> Status:
        """Create a Status object that represents a generic failure."""
        return Status(ErrorCode.FAILURE, message)

    @staticmethod
    def coded_fail(code: ErrorCode, message: str = "") -> Status:
        """Create a Status object that represents a failure."""
        return Status(code, message)


class TypeNotRecoveredError(Exception):
    """
    We tried to recover (i.e. look up in the debugger and craft ourselves) the
    type `name` but failed because of `msg`.
    """

    def __init__(self, name: str, msg: str) -> None:
        self.name = name
        super().__init__(msg)


class TypeNotFoundError(Exception):
    """
    The type is not in the debugger.
    """


class SymbolNotRecoveredError(Exception):
    """
    We tried to recover (i.e. look up in the debugger and find through heuristics) the
    symbol `name` but failed because of `msg`.
    """

    def __init__(self, name: str, msg: str) -> None:
        self.name = name
        super().__init__(msg)



"""
Reading register value from the inferior, and provides a
standardized interface to registers like "sp" and "pc".
"""

from __future__ import annotations

import itertools
from collections import OrderedDict
from collections import defaultdict
from collections.abc import Iterator
from dataclasses import dataclass
from typing import Protocol

from typing_extensions import override

import pwndbg.lib.disasm.helpers as bit_math
from pwndbg.lib.arch import PWNDBG_SUPPORTED_ARCHITECTURES_TYPE

# The printing logic for registers uses the Visitor Pattern
# An implementation of RegisterContextProtocol is defined outside of this class
# (this is a lib/ file, so it shouldn't directly be able to access the process)
#
# Instances of VisitableRegister will call the methods of RegisterContextProtocol to do their logic.


class RegisterContextProtocol(Protocol):
    def flag_register_context(self, reg: str, bit_flags: BitFlags) -> str | None: ...

    def addressing_register_context(self, reg: str, is_virtual: bool) -> str | None: ...

    def segment_registers_context(self, regs: list[str]) -> str | None: ...


# Represents a register or a set of registers that can be printed in the context register view
class VisitableRegister(Protocol):
    def context(self, rc: RegisterContextProtocol) -> str | None: ...


class BitFlags(VisitableRegister):
    # this is intentionally uninitialized -- arm uses the same self.flags structuture for different registers
    # for example
    #   - aarch64_cpsr_flags is used for "cpsr", "spsr_el1", "spsr_el2", "spsr_el3"
    #   - aarch64_sctlr_flags is used for "sctlr", "sctlr_el2", "sctlr_el3"
    regname: str
    flags: OrderedDict[str, int | tuple[int, int]]
    value: int

    def __init__(self, flags: list[tuple[str, int | tuple[int, int]]] = []):
        self.regname = ""
        self.flags = OrderedDict()
        for name, bits in flags:
            self.flags[name] = bits
        self.value = 0

    def __getattr__(self, name: str):
        return getattr(self.flags, name)

    def __getitem__(self, key: str) -> int:
        r = self.flags[key]
        if isinstance(r, int):
            return (self.value >> r) & 1
        s, e = r
        return ((~((1 << s) - 1) & ((1 << (e + 1)) - 1)) & self.value) >> s

    def __setitem__(self, key: str, value: int) -> None:
        self.flags[key] = value

    def __delitem__(self, key: str):
        del self.flags[key]

    def __iter__(self):
        return iter(self.flags)

    def __len__(self):
        return len(self.flags)

    def __repr__(self):
        return f"BitFlags({self.flags})"

    def update(self, regname: str) -> None:
        self.regname = regname

    @override
    def context(self, rc: RegisterContextProtocol) -> str | None:
        return rc.flag_register_context(self.regname, self)


class AddressingRegister(VisitableRegister):
    """
    Represents a register that is used to store an address, e.g. cr3, gsbase, fsbase
    """

    reg: str
    value: int
    is_virtual: bool

    def __init__(self, reg: str, is_virtual: bool):
        self.reg = reg
        self.value = 0
        self.is_virtual = is_virtual

    def update(self, regname: str) -> None:
        pass

    @override
    def context(self, rc: RegisterContextProtocol) -> str | None:
        return rc.addressing_register_context(self.reg, self.is_virtual)


class SegmentRegisters(VisitableRegister):
    """
    Represents the x86 segment register set
    """

    regs: list[str]

    def __init__(self, regs: list[str]):
        self.regs = regs

    @override
    def context(self, rc: RegisterContextProtocol) -> str | None:
        return rc.segment_registers_context(self.regs)


class KernelRegisterSet:
    """
    additional registers that are useful when pwning kernels
    used only for x86-64 for now
    """

    # Segment registers (CS, DS, ES, FS, GS, SS)
    segments: SegmentRegisters

    # Control registers (cr0, cr3, cr4)
    controls: dict[str, BitFlags | AddressingRegister]

    # Model specific registers
    msrs: dict[str, BitFlags | AddressingRegister]

    def __init__(
        self,
        segments: SegmentRegisters,
        controls: dict[str, BitFlags | AddressingRegister] = {},
        msrs: dict[str, BitFlags | AddressingRegister] = {},
    ):
        self.segments = segments
        self.controls = controls
        self.msrs = msrs


@dataclass
class UnicornRegisterWrite:
    """
    Represent a register to write to the Unicorn emulator.
    """

    name: str
    force_write: bool


@dataclass
class Reg:
    name: str
    size: int | None = None
    """Register width in bytes. None if the register size is arch.ptrsize"""
    offset: int = 0
    """Relevant for subregisters - the offset of this register in the main register"""
    zero_extend_writes: bool = False
    """Upon writing a value to this subregister, are the higher bits of the full register zeroed out?"""
    subregisters: tuple[Reg, ...] = ()
    """Bitmask for register. None if the register size is arch.ptrsize"""
    mask: int | None = None

    def __post_init__(self) -> None:
        if self.size:
            self.mask = (1 << (self.size * 8)) - 1


class RegisterSet:
    pc: str
    """Program counter register"""

    stack: str
    """Stack pointer register"""

    frame: str | None = None
    """Frame pointer register"""

    retaddr: tuple[str, ...]
    """Return address register"""

    flags: dict[str, BitFlags]
    """Maps name of flag register (eflags, cpsr) to a structure detailing what the bits mean"""

    gpr: tuple[str, ...]
    """List of native-size general-purpose registers"""

    misc: tuple[str, ...]
    """List of miscellaneous, valid registers"""

    args: tuple[str, ...]
    """Register-based arguments for most common ABI"""

    retval: str | None
    """Return value register"""

    common: list[str] = []
    """Common registers which should be displayed in the register context"""

    kernel: KernelRegisterSet | None
    """Extra registers for kernel debugging"""

    all: set[str]
    """All valid registers"""

    reg_definitions: dict[str, Reg]
    """Map of register name to Reg objects containing information on the register"""

    full_register_lookup: dict[str, Reg]
    """
    Map of register name to the full register it resides in.
    Example mapping: "eax" -> Reg("rax")

    A full size register maps to itself.
    """

    special_aliases: dict[str, str]
    """
    Contains two values:
    - "sp" -> stack pointer register name
    - "pc" -> instruction pointer register name
    """

    def __init__(
        self,
        pc: Reg = Reg("pc"),
        stack: Reg = Reg("sp"),
        frame: Reg | None = None,
        retaddr: tuple[Reg, ...] = (),
        flags: dict[str, BitFlags] = {},
        extra_flags: dict[str, BitFlags] = {},
        gpr: tuple[Reg, ...] = (),
        misc: tuple[str, ...] = (),
        args: tuple[str, ...] = (),
        kernel: KernelRegisterSet | None = None,
        retval: str | None = None,
    ) -> None:
        self.pc = pc.name
        self.stack = stack.name
        self.frame = frame.name if frame else None
        self.retaddr = tuple(x.name for x in retaddr)
        self.flags = flags
        self.extra_flags = extra_flags
        self.gpr = tuple(x.name for x in gpr)
        self.misc = misc
        self.args = args
        self.retval = retval
        self.kernel = kernel

        all_subregisters: list[str] = []

        self.reg_definitions = {}
        self.full_register_lookup = {}
        for reg in itertools.chain(gpr, (stack, frame, pc), retaddr):
            if reg:
                self.reg_definitions[reg.name] = reg
                self.full_register_lookup[reg.name] = reg
                for subregister in reg.subregisters:
                    self.reg_definitions[subregister.name] = subregister
                    self.full_register_lookup[subregister.name] = reg
                    all_subregisters.append(subregister.name)

        # In 'common', we don't want to lose the ordering of:
        self.common = []
        for regname in itertools.chain(
            self.gpr, (self.frame, self.stack, self.pc), tuple(self.flags)
        ):
            if regname and regname not in self.common:
                self.common.append(regname)

        if self.kernel is not None:
            controls = self.kernel.controls
            segments = self.kernel.segments
            msrs = self.kernel.msrs
            for regname in itertools.chain(controls, segments.regs, msrs):
                if regname and regname not in self.common:
                    self.common.append(regname)

        # The specific order of this list is very important:
        # Due to the behavior of Arm in the Unicorn engine,
        # we must write the flags register after PC, and the stack pointer after the flags register.
        # Otherwise, the values will be clobbered
        # https://github.com/pwndbg/pwndbg/pull/2337
        self.emulated_regs_order: list[UnicornRegisterWrite] = []

        # Avoid duplicates
        seen_emulated_register: set[str] = set()

        for regname in itertools.chain(
            (self.pc,),
            tuple(self.flags),
            (self.stack, self.frame),
            self.retaddr,
            self.misc,
            self.gpr,
        ):
            if regname and regname not in seen_emulated_register:
                emu_reg = UnicornRegisterWrite(regname, True if regname in flags else False)
                self.emulated_regs_order.append(emu_reg)
                seen_emulated_register.add(regname)

        self.all = (
            set(self.misc)
            | set(self.flags)
            | set(self.extra_flags)
            | set(self.retaddr)
            | set(self.common)
            | set(all_subregisters)
        )
        self.all -= {None}
        self.all |= {"pc", "sp"}

        self.special_aliases = {}
        self.special_aliases["sp"] = self.stack
        self.special_aliases["pc"] = self.pc

    def resolve_aliases(self, reg: str) -> str:
        """
        Convert "sp" and "pc" to the real architectural registers.
        For all others, returns `reg`
        """
        return self.special_aliases.get(reg, reg)

    def __contains__(self, reg: str) -> bool:
        return reg in self.all

    def __iter__(self) -> Iterator[str]:
        yield from self.all


class PseudoEmulatedRegisterFile:
    """
    This class represents a set of registers that can be written, read, and invalidated.

    The aim is to allow some manual dynamic/static analysis without the need for a full emulator.

    The implementation can handle the behavior of architectures with partial registers,
    such as x86 (Ex: rax has "eax", "ax", "ah", and "al" as subregisters) or AArch64 (Ex: X0 contains W0).
    Most of the complexity of the bitshifts and masks arise from the necessity to handle these cases.
    """

    masks: defaultdict[str, int]
    """
    Map of register name to bitmask indicating what bits of the register we know the value of.

    Example:
    {
        "rax": 0xFFFF
    }
    This indicates that in the RAX register, we only know the bottom 16 bits. This likely resulted from writing the "ax" register.
    Any attempt to read any other bits returns None. In this case, we can read from "ax", "ah", and "al", but not "eax" or "rax".
    """

    values: defaultdict[str, int]
    """
    Map of register to the value we know it to have.
    """

    register_set: RegisterSet
    ptrsize: int

    def __init__(self, register_set: RegisterSet, ptrsize: int):
        self.register_set = register_set
        self.ptrsize = ptrsize

        self.masks = defaultdict(int)
        self.values = defaultdict(int)

    def write_register(
        self, reg: str, value: int, source_width: int | None = None, sign_extend: bool = False
    ) -> None:
        """
        source_width is the byte width of the value's source.
        It should be specified when the source has a width shorter than the destination register.

        Examples:
            movsbl EAX, AL      // sign extend 1 byte register to 4 byte register
            movzbl EAX, AL      // zero extend

            Source width would be 1, and in the first case sign_extend should be set to True.
            If sign_extend is False, we zero extend.
        """
        # Definition of the register we are writing
        write_reg_def = self.register_set.reg_definitions.get(reg)
        if write_reg_def is None:
            return None

        register_bit_offset = write_reg_def.offset * 8
        written_register_size = (
            write_reg_def.size if write_reg_def.size is not None else self.ptrsize
        )
        written_register_mask = (1 << (written_register_size * 8)) - 1

        # Definition of the "full" register that the written register resides in. Might be itself.
        full_reg_def = self.register_set.full_register_lookup[reg]

        # Handle zero / sign-extension
        if source_width is not None:
            # Ensure that if value is negative, it is converted to it's unsigned representation
            value &= (1 << (source_width * 8)) - 1

            # Sign-extend the value to the write_size
            if sign_extend:
                value = bit_math.to_signed(value, source_width * 8) & written_register_mask

        # Bitmask of the register positioned in the full register. Ex: ah register is bits [15-8] in RAX.
        value_mask = written_register_mask << register_bit_offset

        # The bits we will place into the register
        written_bits = (value << register_bit_offset) & value_mask

        if write_reg_def.zero_extend_writes:
            full_reg_size = full_reg_def.size if full_reg_def.size is not None else self.ptrsize
            full_reg_mask = (1 << (full_reg_size * 8)) - 1
            # Bitmask indicating the bits that this write is setting.
            overriden_bits_mask = full_reg_mask
        else:
            overriden_bits_mask = value_mask

        # Clear bits of current value where new value is being written.
        value_masked_for_placement = self.values[full_reg_def.name] & ~overriden_bits_mask

        self.masks[full_reg_def.name] = overriden_bits_mask | self.masks[full_reg_def.name]
        self.values[full_reg_def.name] = written_bits | value_masked_for_placement

    def read_register(self, reg: str) -> int | None:
        # Definition of the register we are reading
        write_reg_def = self.register_set.reg_definitions.get(reg)
        if write_reg_def is None:
            return None

        register_bit_offset = write_reg_def.offset * 8
        written_register_size = (
            write_reg_def.size if write_reg_def.size is not None else self.ptrsize
        )
        written_register_mask = (1 << (written_register_size * 8)) - 1

        # Definition of the "full" register that the read register resides in. Might be itself.
        full_reg_def = self.register_set.full_register_lookup[reg]

        mask = self.masks[full_reg_def.name]
        if mask == 0:
            return None

        read_mask = written_register_mask << register_bit_offset

        if mask & read_mask != read_mask:
            # Not all of the bits that we are attempting to read are readable.
            return None

        return (self.values[full_reg_def.name] & read_mask) >> register_bit_offset

    def invalidate_all_registers(self) -> None:
        self.masks.clear()

    def invalidate_register(self, reg: str) -> None:
        """
        Invalidate the bits that a write to this register would override.

        This can be used when we statically detect that a register is written, but
        we don't know the concrete value that is written so we have to invalidate any current
        knowledge of the register's bits.
        """
        # Definition of the register we are invalidating
        written_reg_def = self.register_set.reg_definitions.get(reg)
        if written_reg_def is None:
            return None

        register_bit_offset = written_reg_def.offset * 8
        written_register_size = (
            written_reg_def.size if written_reg_def.size is not None else self.ptrsize
        )
        written_register_mask = (1 << (written_register_size * 8)) - 1

        # Definition of the "full" register that the written register resides in. Might be itself.
        full_reg_def = self.register_set.full_register_lookup[reg]

        value_mask = written_register_mask << register_bit_offset

        if written_reg_def.zero_extend_writes:
            full_reg_size = full_reg_def.size if full_reg_def.size is not None else self.ptrsize
            full_reg_mask = (1 << (full_reg_size * 8)) - 1
            new_mask = full_reg_mask
        else:
            new_mask = value_mask

        self.masks[full_reg_def.name] = ~new_mask & self.masks[full_reg_def.name]

    def __repr__(self) -> str:
        return str(
            {
                "masks": {x: hex(y) for x, y in self.masks.items()},
                "values": {x: hex(y) for x, y in self.values.items()},
            }
        )


arm_cpsr_flags = BitFlags(
    [
        ("N", 31),
        ("Z", 30),
        ("C", 29),
        ("V", 28),
        ("Q", 27),
        ("J", 24),
        ("T", 5),
        ("E", 9),
        ("A", 8),
        ("I", 7),
        ("F", 6),
    ]
)
arm_xpsr_flags = BitFlags([("N", 31), ("Z", 30), ("C", 29), ("V", 28), ("Q", 27), ("T", 24)])

aarch64_cpsr_flags = BitFlags(
    [
        ("N", 31),
        ("Z", 30),
        ("C", 29),
        ("V", 28),
        ("Q", 27),
        ("PAN", 22),
        ("IL", 20),
        ("D", 9),
        ("A", 8),
        ("I", 7),
        ("F", 6),
        ("EL", 2),
        ("SP", 0),
    ]
)

aarch64_sctlr_flags = BitFlags(
    [
        ("TIDCP", 63),
        ("SPINTMASK", 62),
        ("NMI", 61),
        ("EPAN", 57),
        ("ATA0", 43),
        ("ATA0", 42),
        ("TCF", (40, 2)),
        ("TCF0", (38, 2)),
        ("ITFSB", 37),
        ("BT1", 36),
        ("BT0", 35),
        ("EnIA", 31),
        ("EnIB", 30),
        ("EnDA", 27),
        ("UCI", 26),
        ("EE", 25),
        ("E0E", 24),
        ("SPAN", 23),
        ("TSCXT", 20),
        ("WXN", 19),
        ("nTWE", 18),
        ("nTWI", 16),
        ("UCT", 15),
        ("DZE", 14),
        ("EnDB", 13),
        ("I", 12),
        ("UMA", 9),
        ("SED", 8),
        ("ITD", 7),
        ("nAA", 6),
        ("CP15BEN", 5),
        ("SA0", 4),
        ("SA", 3),
        ("C", 2),
        ("A", 1),
        ("M", 0),
    ]
)

aarch64_tcr_flags = BitFlags(
    [
        ("TG1", (30, 31)),
        ("T1SZ", (16, 21)),
        ("TG0", (14, 15)),
        ("T0SZ", (0, 5)),
    ]
)

aarch64_scr_flags = BitFlags(
    [
        ("HCE", 8),
        ("SMD", 7),
        ("EA", 3),
        ("FIQ", 2),
        ("IRQ", 1),
        ("NS", 0),
    ]
)

aarch64_mmfr_flags = BitFlags([("VARange", (16, 19))])

arm = RegisterSet(
    retaddr=(Reg("lr", 4),),
    flags={"cpsr": arm_cpsr_flags},
    gpr=(
        Reg("r0", 4),
        Reg("r1", 4),
        Reg("r2", 4),
        Reg("r3", 4),
        Reg("r4", 4),
        Reg("r5", 4),
        Reg("r6", 4),
        Reg("r7", 4),
        Reg("r8", 4),
        Reg("r9", 4),
        Reg("r10", 4),
        Reg("r11", 4),
        Reg("r12", 4),
    ),
    args=("r0", "r1", "r2", "r3"),
    retval="r0",
)

# ARM Cortex-M
armcm = RegisterSet(
    retaddr=(Reg("lr", 4),),
    flags={"xpsr": arm_xpsr_flags},
    gpr=(
        Reg("r0", 4),
        Reg("r1", 4),
        Reg("r2", 4),
        Reg("r3", 4),
        Reg("r4", 4),
        Reg("r5", 4),
        Reg("r6", 4),
        Reg("r7", 4),
        Reg("r8", 4),
        Reg("r9", 4),
        Reg("r10", 4),
        Reg("r11", 4),
        Reg("r12", 4),
    ),
    args=("r0", "r1", "r2", "r3"),
    retval="r0",
)

# AArch64 has a PSTATE register, but GDB represents it as the CPSR register
aarch64 = RegisterSet(
    retaddr=(Reg("lr", 8),),  # x30
    flags={"cpsr": aarch64_cpsr_flags},
    extra_flags={
        "scr_el3": aarch64_scr_flags,
        "sctlr": aarch64_sctlr_flags,
        "sctlr_el2": aarch64_sctlr_flags,
        "sctlr_el3": aarch64_sctlr_flags,
        "spsr_el1": aarch64_cpsr_flags,
        "spsr_el2": aarch64_cpsr_flags,
        "spsr_el3": aarch64_cpsr_flags,
        "tcr_el1": aarch64_tcr_flags,
        "id_aa64mmfr2_el1": aarch64_mmfr_flags,
        "ttbr0_el1": BitFlags(),
        "ttbr1_el1": BitFlags(),
    },
    frame=Reg("x29", 8, subregisters=(Reg("w29", 4, zero_extend_writes=True),)),  # x29
    gpr=(
        Reg("x0", 8, subregisters=(Reg("w0", 4, zero_extend_writes=True),)),
        Reg("x1", 8, subregisters=(Reg("w1", 4, zero_extend_writes=True),)),
        Reg("x2", 8, subregisters=(Reg("w2", 4, zero_extend_writes=True),)),
        Reg("x3", 8, subregisters=(Reg("w3", 4, zero_extend_writes=True),)),
        Reg("x4", 8, subregisters=(Reg("w4", 4, zero_extend_writes=True),)),
        Reg("x5", 8, subregisters=(Reg("w5", 4, zero_extend_writes=True),)),
        Reg("x6", 8, subregisters=(Reg("w6", 4, zero_extend_writes=True),)),
        Reg("x7", 8, subregisters=(Reg("w7", 4, zero_extend_writes=True),)),
        Reg("x8", 8, subregisters=(Reg("w8", 4, zero_extend_writes=True),)),
        Reg("x9", 8, subregisters=(Reg("w9", 4, zero_extend_writes=True),)),
        Reg("x10", 8, subregisters=(Reg("w10", 4, zero_extend_writes=True),)),
        Reg("x11", 8, subregisters=(Reg("w11", 4, zero_extend_writes=True),)),
        Reg("x12", 8, subregisters=(Reg("w12", 4, zero_extend_writes=True),)),
        Reg("x13", 8, subregisters=(Reg("w13", 4, zero_extend_writes=True),)),
        Reg("x14", 8, subregisters=(Reg("w14", 4, zero_extend_writes=True),)),
        Reg("x15", 8, subregisters=(Reg("w15", 4, zero_extend_writes=True),)),
        Reg("x16", 8, subregisters=(Reg("w16", 4, zero_extend_writes=True),)),
        Reg("x17", 8, subregisters=(Reg("w17", 4, zero_extend_writes=True),)),
        Reg("x18", 8, subregisters=(Reg("w18", 4, zero_extend_writes=True),)),
        Reg("x19", 8, subregisters=(Reg("w19", 4, zero_extend_writes=True),)),
        Reg("x20", 8, subregisters=(Reg("w20", 4, zero_extend_writes=True),)),
        Reg("x21", 8, subregisters=(Reg("w21", 4, zero_extend_writes=True),)),
        Reg("x22", 8, subregisters=(Reg("w22", 4, zero_extend_writes=True),)),
        Reg("x23", 8, subregisters=(Reg("w23", 4, zero_extend_writes=True),)),
        Reg("x24", 8, subregisters=(Reg("w24", 4, zero_extend_writes=True),)),
        Reg("x25", 8, subregisters=(Reg("w25", 4, zero_extend_writes=True),)),
        Reg("x26", 8, subregisters=(Reg("w26", 4, zero_extend_writes=True),)),
        Reg("x27", 8, subregisters=(Reg("w27", 4, zero_extend_writes=True),)),
        Reg("x28", 8, subregisters=(Reg("w28", 4, zero_extend_writes=True),)),
        # Note: x29 is FP (frame) and x30 is LR (retaddr) register
    ),
    args=("x0", "x1", "x2", "x3"),
    retval="x0",
)


x86flags = {
    "eflags": BitFlags(
        [
            ("CF", 0),
            ("PF", 2),
            ("AF", 4),
            ("ZF", 6),
            ("SF", 7),
            ("IF", 9),
            ("DF", 10),
            ("OF", 11),
            ("AC", 18),
        ]
    )
}

amd64_kernel = KernelRegisterSet(
    segments=SegmentRegisters(["cs", "ss", "ds", "es", "fs", "gs"]),
    controls={
        # only displays the security related bits, otherwise it can be too clustered
        "cr0": BitFlags([("PE", 0), ("WP", 16), ("PG", 31)]),
        "cr3": AddressingRegister("cr3", False),
        "cr4": BitFlags(
            [
                ("UMIP", 11),
                ("FSGSBASE", 16),
                ("SMEP", 20),
                ("SMAP", 21),
                ("PKE", 22),
                ("CET", 23),
                ("PKS", 24),
            ]
        ),
    },
    msrs={
        "efer": BitFlags([("NXE", 11)]),
        "gs_base": AddressingRegister("gs_base", True),
        "fs_base": AddressingRegister("fs_base", True),
    },
)

amd64 = RegisterSet(
    pc=Reg("rip"),
    stack=Reg(
        "rsp",
        8,
        subregisters=(Reg("esp", 4, 0, zero_extend_writes=True), Reg("sp", 2, 0), Reg("spl", 1, 0)),
    ),
    frame=Reg(
        "rbp",
        8,
        subregisters=(Reg("ebp", 4, 0, zero_extend_writes=True), Reg("bp", 2, 0), Reg("bpl", 1, 0)),
    ),
    flags=x86flags,
    gpr=(
        Reg(
            "rax",
            8,
            subregisters=(
                Reg("eax", 4, 0, zero_extend_writes=True),
                Reg("ax", 2, 0),
                Reg("ah", 1, 1),
                Reg("al", 1, 0),
            ),
        ),
        Reg(
            "rbx",
            8,
            subregisters=(
                Reg("ebx", 4, 0, zero_extend_writes=True),
                Reg("bx", 2, 0),
                Reg("bh", 1, 1),
                Reg("bl", 1, 0),
            ),
        ),
        Reg(
            "rcx",
            8,
            subregisters=(
                Reg("ecx", 4, 0, zero_extend_writes=True),
                Reg("cx", 2, 0),
                Reg("ch", 1, 1),
                Reg("cl", 1, 0),
            ),
        ),
        Reg(
            "rdx",
            8,
            subregisters=(
                Reg("edx", 4, 0, zero_extend_writes=True),
                Reg("dx", 2, 0),
                Reg("dh", 1, 1),
                Reg("dl", 1, 0),
            ),
        ),
        Reg(
            "rdi",
            8,
            subregisters=(
                Reg("edi", 4, 0, zero_extend_writes=True),
                Reg("di", 2, 0),
                Reg("dil", 1, 0),
            ),
        ),
        Reg(
            "rsi",
            8,
            subregisters=(
                Reg("esi", 4, 0, zero_extend_writes=True),
                Reg("si", 2, 0),
                Reg("sil", 1, 0),
            ),
        ),
        Reg(
            "r8",
            8,
            subregisters=(
                Reg("r8d", 4, 0, zero_extend_writes=True),
                Reg("r8w", 2, 0),
                Reg("r8b", 1, 0),
            ),
        ),
        Reg(
            "r9",
            8,
            subregisters=(
                Reg("r9d", 4, 0, zero_extend_writes=True),
                Reg("r9w", 2, 0),
                Reg("r9b", 1, 0),
            ),
        ),
        Reg(
            "r10",
            8,
            subregisters=(
                Reg("r10d", 4, 0, zero_extend_writes=True),
                Reg("r10w", 2, 0),
                Reg("r10b", 1, 0),
            ),
        ),
        Reg(
            "r11",
            8,
            subregisters=(
                Reg("r11d", 4, 0, zero_extend_writes=True),
                Reg("r11w", 2, 0),
                Reg("r11b", 1, 0),
            ),
        ),
        Reg(
            "r12",
            8,
            subregisters=(
                Reg("r12d", 4, 0, zero_extend_writes=True),
                Reg("r12w", 2, 0),
                Reg("r12b", 1, 0),
            ),
        ),
        Reg(
            "r13",
            8,
            subregisters=(
                Reg("r13d", 4, 0, zero_extend_writes=True),
                Reg("r13w", 2, 0),
                Reg("r13b", 1, 0),
            ),
        ),
        Reg(
            "r14",
            8,
            subregisters=(
                Reg("r14d", 4, 0, zero_extend_writes=True),
                Reg("r14w", 2, 0),
                Reg("r14b", 1, 0),
            ),
        ),
        Reg(
            "r15",
            8,
            subregisters=(
                Reg("r15d", 4, 0, zero_extend_writes=True),
                Reg("r15w", 2, 0),
                Reg("r15b", 1, 0),
            ),
        ),
    ),
    misc=(
        "cs",
        "ss",
        "ds",
        "es",
        "fs",
        "gs",
        "fs_base",
        "gs_base",
        "ip",
    ),
    kernel=amd64_kernel,
    args=("rdi", "rsi", "rdx", "rcx", "r8", "r9"),
    retval="rax",
)

i386 = RegisterSet(
    pc=Reg("eip"),
    stack=Reg("esp", 4, subregisters=(Reg("sp", 2, 0),)),
    frame=Reg("ebp", 4, subregisters=(Reg("bp", 2, 0),)),
    flags=x86flags,
    gpr=(
        Reg(
            "eax",
            4,
            subregisters=(Reg("ax", 2, 0), Reg("ah", 1, 1), Reg("al", 1, 0)),
        ),
        Reg(
            "ebx",
            4,
            subregisters=(Reg("bx", 2, 0), Reg("bh", 1, 1), Reg("bl", 1, 0)),
        ),
        Reg(
            "ecx",
            4,
            subregisters=(Reg("cx", 2, 0), Reg("ch", 1, 1), Reg("cl", 1, 0)),
        ),
        Reg(
            "edx",
            4,
            subregisters=(Reg("dx", 2, 0), Reg("dh", 1, 1), Reg("dl", 1, 0)),
        ),
        Reg(
            "edi",
            4,
            subregisters=(Reg("di", 2, 0),),
        ),
        Reg(
            "esi",
            4,
            subregisters=(Reg("si", 2, 0),),
        ),
    ),
    misc=(
        "cs",
        "ss",
        "ds",
        "es",
        "fs",
        "gs",
        "fs_base",
        "gs_base",
        "ip",
    ),
    retval="eax",
)


# http://math-atlas.sourceforge.net/devel/assembly/elfspec_ppc.pdf
# r0      Volatile register which may be modified during function linkage
# r1      Stack frame pointer, always valid
# r2      System-reserved register (points at GOT)
# r3-r4   Volatile registers used for parameter passing and return values
# r5-r10  Volatile registers used for parameter passing
# r11-r12 Volatile registers which may be modified during function linkage
# r13     Small data area pointer register (points to TLS)
# r14-r30 Registers used for local variables
# r31     Used for local variables or "environment pointers"
powerpc = RegisterSet(
    retaddr=(Reg("lr"),),
    flags={"msr": BitFlags(), "xer": BitFlags()},
    gpr=(
        Reg("r0"),
        Reg("r1"),
        Reg("r2"),
        Reg("r3"),
        Reg("r4"),
        Reg("r5"),
        Reg("r6"),
        Reg("r7"),
        Reg("r8"),
        Reg("r9"),
        Reg("r10"),
        Reg("r11"),
        Reg("r12"),
        Reg("r13"),
        Reg("r14"),
        Reg("r15"),
        Reg("r16"),
        Reg("r17"),
        Reg("r18"),
        Reg("r19"),
        Reg("r20"),
        Reg("r21"),
        Reg("r22"),
        Reg("r23"),
        Reg("r24"),
        Reg("r25"),
        Reg("r26"),
        Reg("r27"),
        Reg("r28"),
        Reg("r29"),
        Reg("r30"),
        Reg("r31"),
        Reg("cr"),
        Reg("ctr"),
    ),
    args=("r3", "r4", "r5", "r6", "r7", "r8", "r9", "r10"),
    retval="r3",
)

# http://people.cs.clemson.edu/~mark/sparc/sparc_arch_desc.txt
# http://people.cs.clemson.edu/~mark/subroutines/sparc.html
# https://www.utdallas.edu/~edsha/security/sparcoverflow.htm
#
# http://people.cs.clemson.edu/~mark/sparc/assembly.txt
# ____________________________________
# %g0 == %r0  (always zero)           \
# %g1 == %r1                          | g stands for global
# ...                                 |
# %g7 == %r7                          |
# ____________________________________/
# %o0 == %r8                          \
# ...                                 | o stands for output (note: not 0)
# %o6 == %r14 == %sp (stack ptr)      |
# %o7 == %r15 == for return address   |
# ____________________________________/
# %l0 == %r16                         \
# ...                                 | l stands for local (note: not 1)
# %l7 == %r23                         |
# ____________________________________/
# %i0 == %r24                         \
# ...                                 | i stands for input
# %i6 == %r30 == %fp (frame ptr)      |
# %i7 == %r31 == for return address   |
# ____________________________________/

sparc = RegisterSet(
    stack=Reg("sp"),
    frame=Reg("fp"),
    retaddr=(Reg("i7"),),
    flags={"psr": BitFlags()},
    gpr=(
        Reg("g1"),
        Reg("g2"),
        Reg("g3"),
        Reg("g4"),
        Reg("g5"),
        Reg("g6"),
        Reg("g7"),
        Reg("o0"),
        Reg("o1"),
        Reg("o2"),
        Reg("o3"),
        Reg("o4"),
        Reg("o5"),
        Reg("o7"),
        Reg("l0"),
        Reg("l1"),
        Reg("l2"),
        Reg("l3"),
        Reg("l4"),
        Reg("l5"),
        Reg("l6"),
        Reg("l7"),
        Reg("i0"),
        Reg("i1"),
        Reg("i2"),
        Reg("i3"),
        Reg("i4"),
        Reg("i5"),
    ),
    args=("i0", "i1", "i2", "i3", "i4", "i5"),
    retval="o0",
)

# http://logos.cs.uic.edu/366/notes/mips%20quick%20tutorial.htm
# r0        => zero
# r1        => temporary
# r2-r3     => values
# r4-r7     => arguments
# r8-r15    => temporary
# r16-r23   => saved values
# r24-r25   => temporary
# r26-r27   => interrupt/trap handler
# r28       => global pointer
# r29       => stack pointer
# r30       => frame pointer
# r31       => return address
mips = RegisterSet(
    frame=Reg("fp"),
    retaddr=(Reg("ra"),),
    gpr=(
        Reg("v0"),
        Reg("v1"),
        Reg("a0"),
        Reg("a1"),
        Reg("a2"),
        Reg("a3"),
        Reg("t0"),
        Reg("t1"),
        Reg("t2"),
        Reg("t3"),
        Reg("t4"),
        Reg("t5"),
        Reg("t6"),
        Reg("t7"),
        Reg("t8"),
        Reg("t9"),
        Reg("s0"),
        Reg("s1"),
        Reg("s2"),
        Reg("s3"),
        Reg("s4"),
        Reg("s5"),
        Reg("s6"),
        Reg("s7"),
        Reg("s8"),
        Reg("gp"),
    ),
    args=("a0", "a1", "a2", "a3"),
    retval="v0",
)

# https://riscv.org/technical/specifications/
# Volume 1, Unprivileged Spec v. 20191213
# Chapter 25 - RISC-V Assembly Programmer’s Handbook
# x0        => zero   (Hard-wired zero)
# x1        => ra     (Return address)
# x2        => sp     (Stack pointer)
# x3        => gp     (Global pointer)
# x4        => tp     (Thread pointer)
# x5        => t0     (Temporary/alternate link register)
# x6–7      => t1–2   (Temporaries)
# x8        => s0/fp  (Saved register/frame pointer)
# x9        => s1     (Saved register)
# x10-11    => a0–1   (Function arguments/return values)
# x12–17    => a2–7   (Function arguments)
# x18–27    => s2–11  (Saved registers)
# x28–31    => t3–6   (Temporaries)
# f0–7      => ft0–7  (FP temporaries)
# f8–9      => fs0–1  (FP saved registers)
# f10–11    => fa0–1  (FP arguments/return values)
# f12–17    => fa2–7  (FP arguments)
# f18–27    => fs2–11 (FP saved registers)
# f28–31    => ft8–11 (FP temporaries)
riscv = RegisterSet(
    pc=Reg("pc"),
    stack=Reg("sp"),
    retaddr=(Reg("ra"),),
    gpr=(
        Reg("gp"),
        Reg("tp"),
        Reg("t0"),
        Reg("t1"),
        Reg("t2"),
        Reg("s0"),
        Reg("s1"),
        Reg("a0"),
        Reg("a1"),
        Reg("a2"),
        Reg("a3"),
        Reg("a4"),
        Reg("a5"),
        Reg("a6"),
        Reg("a7"),
        Reg("s2"),
        Reg("s3"),
        Reg("s4"),
        Reg("s5"),
        Reg("s6"),
        Reg("s7"),
        Reg("s8"),
        Reg("s9"),
        Reg("s10"),
        Reg("s11"),
        Reg("t3"),
        Reg("t4"),
        Reg("t5"),
        Reg("t6"),
    ),
    args=("a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7"),
    # TODO: make retval a tuple
    # a1 for second return value
    retval="a0",
)

# https://docs.kernel.org/arch/loongarch/introduction.html
loongarch64 = RegisterSet(
    pc=Reg("pc"),
    stack=Reg("sp"),
    frame=Reg("fp"),
    retaddr=(Reg("ra"),),
    gpr=(
        Reg("a0"),
        Reg("a1"),
        Reg("a2"),
        Reg("a3"),
        Reg("a4"),
        Reg("a5"),
        Reg("a6"),
        Reg("a7"),
        Reg("t0"),
        Reg("t1"),
        Reg("t2"),
        Reg("t3"),
        Reg("t4"),
        Reg("t5"),
        Reg("t6"),
        Reg("t7"),
        Reg("t8"),
        Reg("s0"),
        Reg("s1"),
        Reg("s2"),
        Reg("s3"),
        Reg("s4"),
        Reg("s5"),
        Reg("s6"),
        Reg("s7"),
        Reg("s8"),
    ),
    args=(
        "a0",
        "a1",
        "a2",
        "a3",
        "a4",
        "a5",
        "a6",
        "a7",
    ),
    # r21 stores "percpu base address", referred to as "u0" in the kernel
    misc=("tp", "r21"),
)

# https://refspecs.linuxfoundation.org/ELF/zSeries/lzsabi0_zSeries/x410.html
# Register name | Usage                          | Call effect
# --------------|--------------------------------|----------------
# r0            | General purpose               | Volatile
# r1            | General purpose               | Volatile
# r2            | Parameter passing and return  | Volatile
# r3, r4, r5    | Parameter passing             | Volatile
# r6            | Parameter passing             | Saved
# r7 - r11      | Local variables               | Saved
# r12           | Local variable, commonly used | Saved
#               | as GOT pointer                |
# r13           | Local variable, commonly used | Saved
#               | as Literal Pool pointer       |
# r14           | Return address                | Volatile
# r15           | Stack pointer                 | Saved
s390x = RegisterSet(
    pc=Reg("pc"),
    retaddr=(Reg("r14"),),
    stack=Reg("r15"),
    flags={"pswm": BitFlags()},
    gpr=(
        Reg("r0"),
        Reg("r1"),
        Reg("r2"),
        Reg("r3"),
        Reg("r4"),
        Reg("r5"),
        Reg("r6"),
        Reg("r7"),
        Reg("r8"),
        Reg("r9"),
        Reg("r10"),
        Reg("r11"),
        Reg("r12"),
        Reg("r13"),
    ),
    args=("r2", "r3", "r4", "r5", "r6"),
    retval="r2",
)

reg_sets: dict[PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, RegisterSet] = {
    "i386": i386,
    "i8086": i386,
    "x86-64": amd64,
    "rv32": riscv,
    "rv64": riscv,
    "mips": mips,
    "sparc": sparc,
    "arm": arm,
    "armcm": armcm,
    "aarch64": aarch64,
    "powerpc": powerpc,
    "loongarch64": loongarch64,
    "s390x": s390x,
}



"""
Logging.
"""

from __future__ import annotations

import logging

import pwndbg.color.message


class ColorFormatter(logging.Formatter):
    log_funcs = {
        logging.DEBUG: pwndbg.color.message.debug,
        logging.INFO: pwndbg.color.message.info,
        logging.WARNING: pwndbg.color.message.warn,
        logging.ERROR: pwndbg.color.message.error,
        logging.CRITICAL: pwndbg.color.message.error,
    }

    def format(self, record):
        log_func = self.log_funcs.get(record.levelno)
        formatter = logging.Formatter(log_func("%(message)s"))
        return formatter.format(record)



from __future__ import annotations

import pathlib

import pwnlib.context
import pwnlib.data

import pwndbg.aglib
import pwndbg.lib.zig


def _get_pwntools_includes() -> list[pathlib.Path]:
    include = (
        pathlib.Path(pwnlib.data.path)
        / "includes"
        / str(pwnlib.context.context.os)
        / f"{pwnlib.context.context.arch}.h"
    )
    if not include.exists():
        return []
    return [include]


def asm(data: str) -> bytes:
    """
    Assemble the `data` string for the current architecture and return the assembled bytes.

    Only call this if the pwndbg.aglib.arch is not None.
    """
    return pwndbg.lib.zig.asm(pwndbg.aglib.arch, data, includes=_get_pwntools_includes())



"""
Information about whether the debuggee is local (under GDB) or remote
(under GDBSERVER or QEMU stub).
"""

from __future__ import annotations

import pwndbg
import pwndbg.lib.cache


@pwndbg.lib.cache.cache_until("start")
def is_remote() -> bool:
    return pwndbg.dbg.selected_inferior().is_remote()



from __future__ import annotations

import argparse
import ast
import functools
import logging
import math
import re
import sys
from collections import defaultdict
from collections.abc import Callable
from enum import Enum
from typing import Any
from typing import TextIO
from typing import TypeVar

import unicorn as U
from typing_extensions import ParamSpec
from typing_extensions import override

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.disasm.disassembly
import pwndbg.aglib.kernel
import pwndbg.aglib.nearpc
import pwndbg.aglib.qemu
import pwndbg.aglib.signal
import pwndbg.aglib.symbol
import pwndbg.arguments
import pwndbg.chain
import pwndbg.color
import pwndbg.color.context as ctx_color
import pwndbg.color.memory as mem_color
import pwndbg.color.syntax_highlight as H
import pwndbg.commands
import pwndbg.commands.telescope
import pwndbg.dbg_mod
import pwndbg.dintegration
import pwndbg.lib.cache
import pwndbg.lib.config
import pwndbg.lib.pretty_print as pretty_print
import pwndbg.ui
from pwndbg.aglib.arch_mod import get_thumb_mode_string
from pwndbg.color import ColorConfig
from pwndbg.color import ColorParamSpec
from pwndbg.color import message
from pwndbg.color import theme
from pwndbg.commands import CommandCategory
from pwndbg.dbg_mod import EventHandlerPriority
from pwndbg.dbg_mod import EventType
from pwndbg.lib.regs import BitFlags
from pwndbg.lib.regs import RegisterContextProtocol
from pwndbg.lib.regs import VisitableRegister

if pwndbg.dbg.is_gdblib_available():
    import gdb

    import pwndbg.gdblib.ptmalloc2_tracking
    import pwndbg.gdblib.symbol

log = logging.getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")

theme.add_param("backtrace-prefix", "►", "prefix for current backtrace label")

# TODO: Should namespace be "context.backtrace"?
c = ColorConfig(
    "backtrace",
    [
        ColorParamSpec("prefix", "none", "color for prefix of current backtrace label"),
        ColorParamSpec("address", "none", "color for backtrace (address)"),
        ColorParamSpec("symbol", "none", "color for backtrace (symbol)"),
        ColorParamSpec("frame-label", "none", "color for backtrace (frame label)"),
    ],
)


def clear_screen(out=sys.stdout) -> None:
    """
    Clear the screen by moving the cursor to top-left corner and
    clearing the content. Different terminals may act differently
    """
    ## The ANSI escape codes we use here are described e.g. on:
    # https://en.wikipedia.org/wiki/ANSI_escape_code#CSIsection
    #
    ## To sum up the escape codes used below:
    # \x1b - Escape | Starts all the escape sequences
    # [ - Control Sequence Introducer | Starts most of the useful sequences
    # H - Cursor Position | Moves the cursor to row n, column m (default=1)
    # \x1b - Escape | Starts all the escape sequences
    # <n> J - Erase in Display | Clears part of the screen.
    # If n is 0 (or missing), clear from cursor to end of screen.
    # If n is 1, clear from cursor to beginning of the screen.
    # If n is 2, clear entire screen (and moves cursor to upper left on DOS ANSI.SYS).
    # If n is 3, clear entire screen and delete all lines saved in the
    # scrollback buffer (this feature was added for xterm and is supported
    # by other terminal applications
    out.write("\x1b[H\x1b[2J")


config_reserve_lines = pwndbg.config.add_param(
    "context-reserve-lines",
    "if-ctx-fits",
    "when to reserve lines after the prompt to reduce context shake",
    help_docstring="""
The "if-ctx-fits" setting only reserves lines if the whole context would still fit vertically in the current terminal window.
It doesn't take into account line-wrapping due to insufficient terminal width.
""",  # TODO: maybe it could take into account line-wrapping?
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=["never", "if-ctx-fits", "always"],
)


def reserve_lines_maybe(cmd_lines: int) -> None:
    """
    Scroll the terminal up a few lines to reduce shaking
    when repeatedly printing the context.

    Only do this if the context would still fit on
    the screen.
    """
    if config_reserve_lines == "never":
        return

    # If we are clearing the screen, we already have the maximum number of empty
    # lines below us, no need to scroll.
    if config_reserve_lines == "if-ctx-fits" and config_clear_screen:
        return

    # Could use .get_cmd_window_size() but lldb doesn't provide it
    rows, _ = pwndbg.ui.get_window_size(sys.stdout)
    leftover_lines = rows - cmd_lines
    # Take away one line for the prompt (e.g. pwndbg>) which is printed after the context
    leftover_lines -= 1

    # Need one line for the newline generated by pressing enter
    needed_empty = 1
    # Also need to cover extra lines that are printed before we start calculating/printing
    # the context.
    # Even though they are cleared when config_clear_screen=True, they still
    # cause shake, so we still account for them.
    needed_empty += pwndbg.dbg.pre_ctx_lines

    if not config_clear_screen:
        # Since we are not clearing the screen, we assume the prompt is at the bottom
        # of the window
        if config_reserve_lines == "always" or leftover_lines >= needed_empty:
            # We always write to stdout since that is where the prompt will be
            # ANSII codes
            # \x1b[#S  scroll the terminal up by # lines
            # \x1b[#A  move the cursor up by # lines
            sys.stdout.write(f"\x1b[{needed_empty}S\x1b[{needed_empty}A")
    else:
        assert config_reserve_lines == "always"
        # Since we are clearing the screen, leftover_lines describes actual empty lines.
        leftover_lines = max(leftover_lines, 0)
        if leftover_lines < needed_empty:
            to_scroll = needed_empty - leftover_lines
            sys.stdout.write(f"\x1b[{to_scroll}S\x1b[{to_scroll}A")


config_clear_screen = pwndbg.config.add_param(
    "context-clear-screen", False, "whether to clear the screen before printing the context"
)
config_output = pwndbg.config.add_param(
    "context-output", "stdout", 'where Pwndbg should output ("stdout" or file/tty)'
)
config_context_sections = pwndbg.config.add_param(
    "context-sections",
    "last_signal regs disasm code ghidra stack backtrace expressions threads heap_tracker",
    "which context sections are displayed (controls order)",
)
config_max_threads_display = pwndbg.config.add_param(
    "context-max-threads",
    4,
    "maximum number of threads displayed by the context command",
)
config_backtrace_format = pwndbg.config.add_param(
    "context-backtrace-hex",
    False,
    "whether to use hex for offsets in the backtrace",
)

# Storing output configuration per section
OutputType = str | Callable[[str], None]
outputs: dict[str, OutputType] = {}
output_settings: defaultdict[str, dict[str, Any]] = defaultdict(dict)


@pwndbg.config.trigger(config_context_sections)
def validate_context_sections() -> None:
    valid_values = [
        context.__name__.replace("context_", "") for context in context_sections.values()
    ]

    # If someone tries to set an empty string, we let to do that informing about possible values
    # (so that it is possible to have no context at all)
    if not config_context_sections.value or config_context_sections.value.lower() in (
        "''",
        '""',
        "none",
        "empty",
        "-",
    ):
        config_context_sections.value = ""
        print(
            message.warn(
                f"Sections set to be empty. FYI valid values are: {', '.join(valid_values)}"
            )
        )
        return

    for section in config_context_sections.split():
        if section not in valid_values:
            print(
                message.warn(f"Invalid section: {section}, valid values: {', '.join(valid_values)}")
            )
            print(message.warn("(setting none of them like '' will make sections not appear)"))
            config_context_sections.revert_default()
            return


class StdOutput:
    """A context manager wrapper to give stdout"""

    def __enter__(self) -> TextIO:
        return sys.stdout

    def __exit__(self, *args, **kwargs) -> None:
        pass

    def __hash__(self):
        return hash(sys.stdout)

    def __eq__(self, other) -> bool:
        return isinstance(other, StdOutput)


class FileOutput:
    """A context manager wrapper to reopen files on enter"""

    def __init__(self, *args) -> None:
        self.args = args
        self.handle = None

    def __enter__(self) -> TextIO:
        self.handle = open(*self.args)
        return self.handle

    def __exit__(self, *args, **kwargs) -> None:
        self.handle.close()

    def __hash__(self):
        return hash(self.args)

    def __eq__(self, other):
        return self.args == other.args


class CallOutput:
    """A context manager which calls a function on write"""

    def __init__(self, func: Callable[[str], None]) -> None:
        self.func = func

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs) -> None:
        pass

    def __hash__(self):
        return hash(self.func)

    def __eq__(self, other):
        return self.func == other.func

    def write(self, data) -> None:
        self.func(data)

    def writelines(self, lines_iterable) -> None:
        self.func("".join(lines_iterable))

    def flush(self):
        try:
            return self.func.flush()
        except AttributeError:
            pass

    def isatty(self):
        try:
            return self.func.isatty()
        except AttributeError:
            return False


OutputWrapper = StdOutput | FileOutput | CallOutput
OutputTarget = TextIO | CallOutput


def output(section: str) -> OutputWrapper:
    """Creates a context manager corresponding to configured context output"""
    target = outputs.get(section, str(config_output))
    if not target or target == "stdout":
        return StdOutput()
    if callable(target):
        return CallOutput(target)
    return FileOutput(target, "w")


parser = argparse.ArgumentParser(description="Sets the output of a context section.")
parser.add_argument(
    "section",
    type=str,
    help="The section which is to be configured. ('regs', 'disasm', 'code', 'stack', 'backtrace', 'args', 'threads', 'heap_tracker', 'expressions', and/or 'last_signal')",
)
parser.add_argument("path", type=str, help="The path to which the output is written")
parser.add_argument("clearing", type=bool, help="Indicates whether to clear the output")
banner_arg = parser.add_argument(
    "banner",
    type=str,
    nargs="?",
    default="both",
    help="Where a banner should be placed: both, top , bottom, none",
)
parser.add_argument(
    "width",
    type=int,
    nargs="?",
    default=None,
    help="Sets a fixed width (used for banner). Set to None for auto",
)
parser.add_argument(
    "height",
    type=int,
    nargs="?",
    default=None,
    help="Sets a fixed height (used for number of lines to display before cutoff). Only applies to sections where more data can be displayed like disasm, stack, and threads and overrides their section specific limits. Set to None for auto",
)


@pwndbg.commands.Command(parser, aliases=["ctx-out"], category=CommandCategory.CONTEXT)
def contextoutput(
    section: str,
    path: OutputType,
    clearing: bool,
    banner: str = "both",
    width: int | None = None,
    height: int | None = None,
):
    if not banner:  # synonym for splitmind backwards compatibility
        banner = "none"
    elif banner not in ("both", "top", "bottom", "none"):
        raise argparse.ArgumentError(banner_arg, f"banner can not be '{banner}'")

    outputs[section] = path
    output_settings[section].update(
        {
            "clearing": clearing,
            "width": width,
            "height": height,
            "banner_top": banner in ("both", "top"),
            "banner_bottom": banner in ("both", "bottom"),
        }
    )


def resetcontextoutput(section: str) -> None:
    target = outputs.pop(section, None)
    if target is not None:
        # Remove all settings except for the ones that are not related to output redirection
        output_settings[section] = {
            k: v
            for k, v in output_settings[section].items()
            if k not in ("clearing", "width", "height", "banner_top", "banner_bottom")
        }


# Context history
context_history: defaultdict[str, list[list[str]]] = defaultdict(list)
selected_history_index: int | None = None

context_history_size = pwndbg.config.add_param(
    "context-history-size", 50, "number of context history entries to store"
)


@pwndbg.config.trigger(context_history_size)
def history_size_changed() -> None:
    if context_history_size <= 0:
        context_history.clear()
    else:
        for section in context_history:
            context_history[section] = context_history[section][-int(context_history_size) :]


def serve_context_history(function: Callable[P, list[str]]) -> Callable[P, list[str]]:
    @functools.wraps(function)
    def _serve_context_history(*a: P.args, **kw: P.kwargs) -> list[str]:
        global selected_history_index
        assert "context_" in function.__name__
        section_name = function.__name__.replace("context_", "")

        # If the history is disabled, just return the current output
        if context_history_size <= 0:
            return function(*a, **kw)

        # Add the current section to the history if it is not already there
        current_output = []
        if pwndbg.aglib.proc.alive():
            # Do not reevaluate the expressions section because its content is not deterministic.
            # Instead, reuse the last evaluated expression and rely on the other sections to deselect
            # the history entry if the output changed.
            # https://github.com/pwndbg/pwndbg/issues/2579
            if (
                section_name == "expressions"
                and selected_history_index is not None
                and len(context_history[section_name]) > 0
            ):
                current_output = context_history[section_name][-1]
            else:
                current_output = function(*a, **kw)
            if (
                len(context_history[section_name]) == 0
                or context_history[section_name][-1] != current_output
            ):
                context_history[section_name].append(current_output)
                selected_history_index = None
        # Show the history if the process is not running anymore
        elif context_history[section_name] and selected_history_index is None:
            selected_history_index = len(context_history[section_name]) - 1

        # Truncate the history to the configured size
        context_history[section_name] = context_history[section_name][-int(context_history_size) :]
        history = context_history[section_name]

        if selected_history_index is None:
            return current_output or function(*a, **kw)
        if not history or selected_history_index >= len(history):
            return []
        return history[selected_history_index]

    return _serve_context_history


def history_handle_unchanged_contents() -> None:
    if not context_history:
        return
    longest_history = max(len(h) for h in context_history.values())
    for section_name, history in context_history.items():
        # Duplicate the last entry if it is the same as the previous one
        # and wasn't added when the history was updated
        if len(history) == longest_history - 1:
            context_history[section_name].append(history[-1])
        # Prepend empty entries to the history to make all sections have the same length
        elif len(history) < longest_history - 1:
            context_history[section_name] = [
                [] for _ in range(longest_history - 1 - len(history))
            ] + history


parser = argparse.ArgumentParser(description="Select previous entry in context history.")
parser.add_argument(
    "count",
    type=int,
    nargs="?",
    default=1,
    help="The number of entries to go back in history",
)


@pwndbg.commands.Command(parser, aliases=["ctxp"], category=CommandCategory.CONTEXT)
def contextprev(count: int) -> None:
    global selected_history_index
    if not context_history:
        print(message.error("No context history captured"))
        return
    if selected_history_index is None:
        longest_history = max(len(h) for h in context_history.values())
        new_index = longest_history - count - 1
    else:
        new_index = selected_history_index - count
    selected_history_index = max(0, new_index)
    context()


parser = argparse.ArgumentParser(description="Select next entry in context history.")
parser.add_argument(
    "count",
    type=int,
    nargs="?",
    default=1,
    help="The number of entries to go forward in history",
)


@pwndbg.commands.Command(parser, aliases=["ctxn"], category=CommandCategory.CONTEXT)
def contextnext(count: int) -> None:
    global selected_history_index
    if not context_history:
        print(message.error("No context history captured"))
        return
    longest_history = max(len(h) for h in context_history.values())
    if selected_history_index is None:
        new_index = longest_history - 1
    else:
        new_index = selected_history_index + count
    selected_history_index = min(longest_history - 1, new_index)
    context()


parser = argparse.ArgumentParser(
    description="Search for a string in the context history and select that entry."
)
parser.add_argument(
    "needle",
    type=str,
    help="The string to search for in the context history",
)
parser.add_argument(
    "section",
    type=str,
    nargs="?",
    default=None,
    help="The section to search in. If not provided, search in all sections",
)


@pwndbg.commands.Command(parser, aliases=["ctxsearch"], category=CommandCategory.CONTEXT)
def contextsearch(needle: str, section: str | None = None) -> None:
    if not section:
        sections = context_history.keys()
    else:
        if section not in context_history:
            print(message.error(f"Section '{section}' not found in context history."))
            return
        sections = [section]

    matches: list[tuple[str, int]] = []
    for section in sections:
        for i, entry in enumerate(context_history[section]):
            if not any(m[1] == i for m in matches) and any(needle in line for line in entry):
                matches.append((section, i))
    matches.sort(key=lambda m: m[1], reverse=True)

    if not matches:
        print(message.error(f"String '{needle}' not found in context history."))
        return

    # Select first match before currently selected entry
    global selected_history_index
    if selected_history_index is None:
        next_match = matches[0]
    else:
        for match in matches:
            if match[1] < selected_history_index:
                next_match = match
                break
        else:
            next_match = matches[0]
            print(message.warn("No more matches before the current entry. Starting from the top."))

    selected_history_index = next_match[1]
    print(
        message.info(
            f"Found {len(matches)} match{'es' if len(matches) > 1 else ''}. Selected entry {next_match[1] + 1} for match in section '{next_match[0]}'."
        )
    )
    context()


# Watches
expressions = []

parser = argparse.ArgumentParser(
    description="""
Adds an expression to be shown on context.

To remove an expression, see `cunwatch`.
""",
)
parser.add_argument(
    "cmd",
    type=str,
    default="eval",
    nargs="?",
    choices=("eval", "execute"),
    help="""Command to be used with the expression.
- eval: the expression is parsed and evaluated as in the debugged language.
- execute: the expression is executed as a GDB command.""",
)
parser.add_argument(
    "expression", type=str, help="The expression to be evaluated and shown in context"
)


@pwndbg.commands.Command(
    parser,
    aliases=["ctx-watch", "cwatch"],
    category=CommandCategory.CONTEXT,
    examples="""
For watching variables/expressions:
    cwatch BUF
    cwatch ITEMS[0]

For running commands:
    cwatch execute "ds BUF"
    cwatch execute "x/20x $rsp"
    cwatch execute "info args"
    """,
)
def contextwatch(expression: str, cmd: str = "eval") -> None:
    expressions.append((expression, cmd))


parser = argparse.ArgumentParser(
    description="Removes an expression previously added to be watched."
)
parser.add_argument("num", type=int, help="The expression number to be removed from context")


@pwndbg.commands.Command(
    parser, aliases=["ctx-unwatch", "cunwatch"], category=CommandCategory.CONTEXT
)
def contextunwatch(num: int) -> None:
    if num < 1 or num > len(expressions):
        print(message.error("Invalid input"))
        return

    expressions.pop(int(num) - 1)


@serve_context_history
def context_expressions(
    target: OutputTarget = sys.stdout,
    with_banner: bool = True,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    if not expressions:
        return []
    banner = [pwndbg.ui.banner("expressions", target=target, width=width)]
    output = []
    for i, (exp, cmd) in enumerate(expressions):
        header = f"{i + 1}: {ctx_color.highlight(exp)}"
        try:
            if cmd == "eval":
                value = str(gdb.parse_and_eval(exp))
            else:
                assert cmd == "execute"
                value = gdb.execute(exp, from_tty=False, to_string=True)
        except gdb.error as err:
            value = str(err)

        # When evaluating the expression we display it inline with the header, but when executing an
        # expression we display it on the next line
        if cmd == "eval":
            header += f" = {value}"
        output.append(header)

        if cmd == "execute":
            output.append(value)

    # gdb can output single elements with \n's
    # we split it up here so the cmd_line calculation is correct in context()
    _output = []
    for val in output:
        _output.extend(val.split("\n"))
    output = _output

    return banner + output if with_banner else output


parser = argparse.ArgumentParser(
    description="""
Print out the currently enabled context sections.

This is the text that gets printed on every stop. It can be useful
to run this command manually when you change some process/debugger
state but don't want to step/continue (e.g. after using the `down`
and `up` commands).
"""
)
parser.add_argument(
    "subcontext",
    nargs="*",
    type=str,
    default=None,
    help="Submenu to display: 'regs', 'disasm', 'code', 'stack', 'backtrace', 'args', 'threads', 'heap_tracker', 'expressions', and/or 'last_signal'",
)
parser.add_argument(
    "--on",
    dest="enabled",
    action="store_true",
    default=None,
    help="Show the section(s) in subsequent context commands again. The section(s) have to be in the 'context-sections' list.",
)
parser.add_argument(
    "--off",
    dest="enabled",
    action="store_false",
    default=None,
    help="Do not show the section(s) in subsequent context commands even though they might be in the 'context-sections' list.",
)
parser.add_argument(
    "-a",
    "--all",
    dest="all_sections",
    action="store_true",
    default=False,
    help="Show all context sections.",
)


@pwndbg.commands.Command(
    parser,
    aliases=["ctx"],
    category=CommandCategory.CONTEXT,
    notes="""
To see more commands related to context control run:
```
pwndbg -c context
```
To see context configuration run:
```
config context
```
""",
)
def context(
    subcontext: list[str] | None = None,
    enabled: bool | None = None,
    all_sections: bool = False,
) -> None:
    """
    Print out the current register, instruction, and stack context.

    Accepts subcommands 'reg', 'disasm', 'code', 'stack', 'backtrace', 'args', 'threads', 'heap_tracker', 'expressions', and/or 'last_signal'.
    """
    # Allow to view history after the program has exited
    if not pwndbg.aglib.proc.alive() and (context_history_size <= 0 or not context_history):
        log.error("context: The program is not being run.")
        return None

    if subcontext is None:
        subcontext = []
    args: list[str] = subcontext

    if all_sections:
        args = [c.__name__.replace("context_", "") for c in context_sections.values()]
    elif len(args) == 0:
        args = config_context_sections.split()

    sections: list[tuple[str, Callable[..., list[str]] | None]] = []
    if args:
        if selected_history_index is None:
            sections.append(("legend", lambda *args, **kwargs: [mem_color.legend()]))
        else:
            longest_history = max(len(h) for h in context_history.values())
            history_status = f" (history {selected_history_index + 1}/{longest_history})"
            sections.append(
                ("legend", lambda *args, **kwargs: [mem_color.legend() + history_status])
            )

    sections += [(arg, context_sections.get(arg[0], None)) for arg in args]

    result: defaultdict[OutputWrapper, list[str]] = defaultdict(list)
    result_settings: defaultdict[OutputWrapper, dict[str, Any]] = defaultdict(dict)
    for section, func in sections:
        if func:
            target = output(section)
            # Last section of an output decides about output settings
            settings = output_settings[section]
            if enabled is not None:
                settings["enabled"] = enabled
            if settings.get("enabled", True):
                result_settings[target].update(settings)
                with target as out:
                    result[target].extend(
                        func(
                            target=out,
                            width=settings.get("width", None),
                            height=settings.get("height", None),
                            with_banner=settings.get("banner_top", True),
                        )
                    )

    history_handle_unchanged_contents()

    for target, res in result.items():
        settings = result_settings[target]
        if len(res) > 0 and settings.get("banner_bottom", True):
            with target as out:
                res.append(pwndbg.ui.banner("", target=out, width=settings.get("width", None)))

    cmd_lines = 0
    for target, lines in result.items():
        with target as out:
            if result_settings[target].get("clearing", config_clear_screen) and lines:
                clear_screen(out)
            # The `if` here essentially checks whether the output is going to
            # the window where the debugger prompt is. The prompt is
            # always printed to stdout.
            if isinstance(target, StdOutput):
                # We always get a new line after the prompt, so the first line
                # doesn't need to have a "\n" in front. Furthermore, we need
                # to reserve a new line before the next prompt is printed.
                out.writelines(line + "\n" for line in lines)
            else:
                # A "\n" after the last line would just take up space in this case.
                # There will be a redundant preceding empty line when the first
                # context is printed, but it doesn't really matter.
                out.writelines("\n" + line for line in lines)
            out.flush()
        # Assuming the command window (where the prompt is printed) is always stdout
        if isinstance(target, StdOutput):
            assert all("\n" not in line for line in lines)
            cmd_lines += len(lines)

    reserve_lines_maybe(cmd_lines)


class CompactRegsOptions(Enum):
    NO = "off"
    YES = "on"
    VERY = "very"
    HARDCUT = "hardcut"


pwndbg.config.add_param(
    "show-compact-regs",
    CompactRegsOptions.NO.value,
    "whether to show a compact register view with columns",
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=[x.value for x in CompactRegsOptions],
    help_docstring=f"""
Values explained:

+ `{CompactRegsOptions.NO.value}` - Disable compact registers (default). Every other option tries to make the register context use less rows by putting the registers into multiple columns.
+ `{CompactRegsOptions.YES.value}` - If a register printout doesn't fit it will be added to the end of the register context.
+ `{CompactRegsOptions.VERY.value}` - Try to very hard to compress. May save more lines than `{CompactRegsOptions.YES.value}` but logical register grouping may suffer.
+ `{CompactRegsOptions.HARDCUT.value}` - If a register printout doesn't fit its slot, it will simply be truncated.

See also show-compact-regs-columns, show-compact-regs-min-width and show-compact-regs-separation.
""",
)
pwndbg.config.add_param(
    "show-compact-regs-columns", 2, "the number of columns (0 for dynamic number of columns)"
)
pwndbg.config.add_param("show-compact-regs-min-width", 20, "the minimum width of each column")
pwndbg.config.add_param(
    "show-compact-regs-separation", 4, "the number of spaces separating columns"
)


def calculate_padding_to_align(length: int, align: int) -> int:
    """Calculates the number of spaces to append to reach the next alignment.
    The next alignment point is given by "x * align >= length".
    """
    return 0 if length % align == 0 else (align - (length % align))


def compact_regs_hardcut(
    regs: list[str], terminal_width: int, column_width: int, columns: int, separation: int
) -> list[str]:
    """
    If the string of any register overflows its column_width, it will be hard cut to the column_width.

    Example:
     RAX  0xfffffffffffffdfe              R8   0                               R14  0
     RBX  0                               R9   0x7fffffffcbd0 —▸ 0x7fff...     R15  0x7ffff7f83e60 (_rl_orig...
     RCX  0x7ffff7c90efa (__intern...     R10  0                               RBP  1
     RDX  0                               R11  0x202                           RSP  0x7fffffffcb70 ◂— 0
     RDI  1                               R12  0x7fffffffccb0 ◂— 1             RIP  0x7ffff7c90efa (__intern...
     RSI  0x7fffffffccb0 ◂— 1             R13  0                               EFLAGS 0x202 [ cf pf af zf sf...
    """
    result: list[str] = []

    cut_marker = pwndbg.color.white("...")

    def hardcut(reg: str) -> tuple[str, int]:
        # Returns the cut string and the new size
        # I don't know of a better way to do this while retaining the coloring.
        reglen = len(reg)
        for i in range(reglen, 0, -1):
            candidate = reg[0:i] + cut_marker
            candidate_len = len(pwndbg.color.strip(candidate))
            if candidate_len <= column_width:
                return candidate, candidate_len
        # Shouldn't happen anyway, but we return non-zero so padding alignment
        # can proceed.
        return " ", 1

    line: str = ""
    line_length: int = 0
    nregs: int = len(regs)
    nrows: int = math.ceil(nregs / columns)
    for row_idx in range(nrows):
        for column_idx in range(columns):
            # Pad the line from the last register.
            if column_idx != 0:
                padding = calculate_padding_to_align(line_length, column_width + separation)
                line += " " * padding
                line_length += padding

            reg_idx = column_idx * nrows + row_idx
            if reg_idx >= nregs:
                # Some columns will not have all rows filled.
                continue
            reg = regs[reg_idx]

            # Strip the color / hightlight information the get the raw text width of the register
            reg_length = len(pwndbg.color.strip(reg))

            if reg_length > column_width:
                txt, txtlen = hardcut(reg)
                line += txt
                line_length += txtlen
            else:
                line += reg
                line_length += reg_length

        # Add the line.
        result.append(line)
        line = ""
        line_length = 0

    return result


def compact_regs_normal(
    regs: list[str], terminal_width: int, column_width: int, columns: int, separation: int
) -> list[str]:
    """
    Will try to group similar registers together, and may increase the number of rows (as opposed to compact_regs_very)
    in order to achieve this.

    column_width does not include separation.

    Example:
     RAX  0xfffffffffffffdfe           R8   0                            R14  0
     RBX  0                            R9    ⏎                           R15   ⏎
     RCX   ⏎                           R10  0                            RBP  1
     RDX  0                            R11  0x202                        RSP  0x7fffffffcb70 ◂— 0
     RDI  1                            R12  0x7fffffffccb0 ◂— 1          RIP   ⏎
     RSI  0x7fffffffccb0 ◂— 1          R13  0                            EFLAGS   ⏎
    ↪ R9   0x7fffffffcbd0 —▸ 0x7ffff7f83e60 (_rl_orig_sigset) ◂— 0
    ↪ R15  0x7ffff7f83e60 (_rl_orig_sigset) ◂— 0
    ↪ RCX  0x7ffff7c90efa (__internal_syscall_cancel+138) ◂— add rsp, 0x18
    ↪ RIP  0x7ffff7c90efa (__internal_syscall_cancel+138) ◂— add rsp, 0x18
    ↪ EFLAGS 0x202 [ cf pf af zf sf IF df of ac ]
    """
    result: list[str] = []

    def extract_reg_name(one_reg: str) -> str:
        # We want the whitespace right after the name too. Also the change marker.
        # Tricky because we want to preserve colors.
        # state = 0 means i'm before the register name
        # state = 1 means i'm in the register name
        # state = 2 means i'm in the whitespace after the register name
        state: int = 0
        last_ws: int = -1
        for i in range(len(one_reg)):
            match state:
                case 0:
                    if one_reg[i] == change_marker or one_reg[i] == " ":
                        state = 1
                case 1:
                    if one_reg[i] == " ":
                        state = 2
                case 2:
                    if one_reg[i] != " ":
                        last_ws = i - 1
                        break
        assert last_ws != -1
        return one_reg[:last_ws]

    will_wrap_char: str = pwndbg.color.light_gray("  ⏎")
    wrapping_char: str = pwndbg.color.gray("↪")
    change_marker: str = str(ctx_color.config_register_changed_marker)

    line: str = ""
    line_length: int = 0
    nregs: int = len(regs)
    nrows: int = math.ceil(nregs / columns)
    pending: list[str] = []
    for row_idx in range(nrows):
        for column_idx in range(columns):
            # Pad the line from the last register.
            if column_idx != 0:
                padding = calculate_padding_to_align(line_length, column_width + separation)
                line += " " * padding
                line_length += padding

            reg_idx = column_idx * nrows + row_idx
            if reg_idx >= nregs:
                # Some columns will not have all rows filled.
                continue
            reg = regs[reg_idx]

            # Strip the color / hightlight information the get the raw text width of the register
            reg_length = len(pwndbg.color.strip(reg))

            if reg_length > column_width:
                # We cannot put this register here without displacing the next one, so we will
                # print this register in a lone line at the end.
                pending.append(reg)
                # We want to leave a marker that we will wrap here.
                # We extract the register name from the `reg` string which is a bit tricky.
                # Look at the RegisterContext class for reference.
                reg_name = extract_reg_name(reg)
                addition = reg_name + will_wrap_char
                line += addition
                line_length += len(pwndbg.color.strip(addition))
            else:
                line += reg
                line_length += reg_length

        # Add the line.
        result.append(line)
        line = ""
        line_length = 0

    # Add all registers that couldn't fit into their slot.
    if pending:
        for pending_line in pending:
            result.append(wrapping_char + pending_line)
        pending.clear()

    return result


def compact_regs_very(
    regs: list[str], terminal_width: int, column_width: int, columns: int, separation: int
) -> list[str]:
    """
    Will try to group similar registers together, but will sacrifice the grouping if it can be more compact.

    column_width does not include separation.

    Example:
     RAX  0xfffffffffffffdfe           R8   0                            R14  0
     RBX  0                            R9   0x7fffffffcbd0 —▸ 0x7ffff7f83e60 (_rl_orig_sigset) ◂— 0
     R15  0x7ffff7f83e60 (_rl_orig_sigset) ◂— 0
     RCX  0x7ffff7c90efa (__internal_syscall_cancel+138) ◂— add rsp, 0x18
     R10  0                            RBP  1                            RDX  0
     R11  0x202                        RSP  0x7fffffffcb70 ◂— 0          RDI  1
     R12  0x7fffffffccb0 ◂— 1
     RIP  0x7ffff7c90efa (__internal_syscall_cancel+138) ◂— add rsp, 0x18
     RSI  0x7fffffffccb0 ◂— 1          R13  0
     EFLAGS 0x202 [ cf pf af zf sf IF df of ac ]
    """
    result: list[str] = []

    line: str = ""
    line_length: int = 0
    nregs: int = len(regs)
    nrows: int = math.ceil(nregs / columns)
    for row_idx in range(nrows):
        for column_idx in range(columns):
            reg_idx = column_idx * nrows + row_idx
            if reg_idx >= nregs:
                # Some columns will not have all rows filled.
                continue
            reg = regs[reg_idx]

            # Strip the color / hightlight information the get the raw text width of the register
            reg_length = len(pwndbg.color.strip(reg))

            # Length of line with unoccupied space and padding is required
            # to fit the register string onto the screen / display.
            line_length_with_padding = line_length
            line_length_with_padding += (
                separation if line_length != 0 else 0
            )  # No separation at the start of a line
            line_length_with_padding += calculate_padding_to_align(
                line_length_with_padding, column_width + separation
            )

            # When element does not fully fit, then start a new line
            if line_length_with_padding + max(reg_length, column_width) > terminal_width:
                result.append(line)

                line = ""
                line_length = 0
                line_length_with_padding = 0

            # Add padding in front of the next printed register
            if line_length != 0:
                line += " " * (line_length_with_padding - line_length)

            line += reg
            line_length = line_length_with_padding + reg_length

    # Append last line if required
    if line_length != 0:
        result.append(line)

    return result


def compact_regs(
    regs: list[str], width: int | None = None, target: OutputTarget = sys.stdout
) -> list[str]:
    columns = max(0, int(pwndbg.config.show_compact_regs_columns))
    min_width = max(1, int(pwndbg.config.show_compact_regs_min_width))
    separation = max(1, int(pwndbg.config.show_compact_regs_separation))

    if width is None:
        _, width = pwndbg.ui.get_window_size(target)

    if columns > 0:
        # Adjust the minimum_width (column) according to the
        # layout depicted below, where there are "columns" columns
        # and "columns - 1" separations.
        #
        # |<----------------- window width -------------------->|
        # | column | sep. | column | sep. | ... | sep. | column |
        #
        # Which results in the following formula:
        # window_width = columns * min_width + (columns - 1) * separation
        # => min_width = (window_width - (columns - 1) * separation) / columns
        min_width = max(min_width, (width - (columns - 1) * separation) // columns)

    match pwndbg.config.show_compact_regs.value:
        case CompactRegsOptions.YES.value:
            return compact_regs_normal(regs, width, min_width, columns, separation)
        case CompactRegsOptions.VERY.value:
            return compact_regs_very(regs, width, min_width, columns, separation)
        case CompactRegsOptions.HARDCUT.value:
            return compact_regs_hardcut(regs, width, min_width, columns, separation)
        case _:
            assert False, "Invalid compact regs value."


@serve_context_history
def context_regs(
    target: OutputTarget = sys.stdout,
    with_banner: bool = True,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    regs = get_regs()
    if pwndbg.config.show_compact_regs.value != CompactRegsOptions.NO.value:
        regs = compact_regs(regs, target=target, width=width)

    info = " / show-flags {} / show-compact-regs {}".format(
        "on" if pwndbg.config.show_flags else "off", pwndbg.config.show_compact_regs
    )
    banner = [pwndbg.ui.banner("registers", target=target, width=width, extra=info)]
    return banner + regs if with_banner else regs


@serve_context_history
def context_heap_tracker(
    target: OutputTarget = sys.stdout,
    with_banner: bool = True,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    if not pwndbg.gdblib.ptmalloc2_tracking.is_enabled():
        return []

    banner = [pwndbg.ui.banner("heap tracker", target=target, width=width, extra="")]

    if pwndbg.gdblib.ptmalloc2_tracking.last_issue is not None:
        info = [
            f"Detected the following potential issue: {pwndbg.gdblib.ptmalloc2_tracking.last_issue}"
        ]
        pwndbg.gdblib.ptmalloc2_tracking.last_issue = None
    else:
        info = ["Nothing to report."]

    return banner + info if with_banner else info


parser = argparse.ArgumentParser(description="Print out all registers and enhance the information.")
parser.add_argument("regs", nargs="*", type=str, default=None, help="Registers to be shown")


@pwndbg.commands.Command(parser, category=CommandCategory.CONTEXT)
@pwndbg.commands.OnlyWhenRunning
def regs(regs: list[str | VisitableRegister | None] | None = None) -> None:
    """Print out all registers and enhance the information."""
    print("\n".join(get_regs(regs)))


pwndbg.config.add_param("show-flags", False, "whether to show flags registers")
pwndbg.config.add_param("show-retaddr-reg", True, "whether to show return address register")


class RegisterContext(RegisterContextProtocol):
    changed: list[str]

    def __init__(self):
        self.changed = pwndbg.aglib.regs.changed

    def get_prefix(self, reg: str) -> str:
        # Make the register stand out and give a color if changed
        regname = ctx_color.register(reg.ljust(4).upper())
        if reg in self.changed:
            regname = ctx_color.register_changed(regname)

        # Show a marker next to the register if it changed
        change_marker = f"{ctx_color.config_register_changed_marker}"
        m = (
            " " * len(change_marker)
            if reg not in self.changed
            else ctx_color.register_changed(change_marker)
        )
        return f"{m}{regname}"

    def get_register_value(self, reg: str) -> int | None:
        val = pwndbg.aglib.regs.read_reg(reg)
        if val is None:
            print(message.warn(f"Unknown register: {reg!r}"))
            return None
        return val

    @override
    def flag_register_context(self, reg: str, bit_flags: BitFlags) -> str | None:
        val = self.get_register_value(reg)
        if val is None:
            return None
        desc = ctx_color.format_flags(val, bit_flags, pwndbg.aglib.regs.last.get(reg, 0))
        prefix = self.get_prefix(reg)
        return f"{prefix} {desc}"

    @override
    def segment_registers_context(self, regs: list[str]) -> str | None:
        result = ""
        for reg in regs:
            val = self.get_register_value(reg)
            if val is None:
                continue
            prefix = self.get_prefix(reg)
            result += f"{prefix} {hex(val)}   "
        return result

    @override
    def addressing_register_context(self, reg: str, is_virtual: bool) -> str | None:
        if is_virtual:
            return self.register_context_default(reg)
        val = self.get_register_value(reg)
        if val is None:
            return None
        prefix = self.get_prefix(reg)
        desc = hex(val)
        if pwndbg.aglib.kernel.has_debug_symbols():
            try:
                virtual = pwndbg.aglib.kernel.phys_to_virt(val)
                desc += f" [virtual: {pwndbg.chain.format(virtual)}]"
            except Exception:
                pass
        return f"{prefix} {desc}"

    def register_context_default(self, reg: str) -> str | None:
        val = self.get_register_value(reg)
        if val is None:
            return None
        desc = ""
        desc = pwndbg.chain.format(val)
        prefix = self.get_prefix(reg)
        return f"{prefix} {desc}"


def get_regs(in_regs: list[str | VisitableRegister | None] | None = None) -> list[str]:
    # Python default parameters are instantiated once and shared across calls.
    # Instead of a default value of [], we need to do this check so we get a fresh list each time
    if in_regs is None:
        in_regs = []
    regs: list[str | VisitableRegister | None] = in_regs
    result: list[str] = []
    rc = RegisterContext()

    if len(regs) == 0:
        regs += pwndbg.aglib.regs.gpr

        regs.append(pwndbg.aglib.regs.frame)
        regs.append(pwndbg.aglib.regs.stack)

        if pwndbg.config.show_retaddr_reg:
            regs += pwndbg.aglib.regs.retaddr

        regs.append(pwndbg.aglib.regs.current.pc)

        if pwndbg.aglib.qemu.is_qemu_kernel() and pwndbg.aglib.regs.kernel is not None:
            controls = pwndbg.aglib.regs.kernel.controls
            if controls is not None:
                for regname, control in controls.items():
                    control.update(regname)
                    regs.append(control)
            msrs = pwndbg.aglib.regs.kernel.msrs
            if msrs is not None:
                for regname, msr in msrs.items():
                    msr.update(regname)
                    regs.append(msr)
        if pwndbg.config.show_flags:
            flags = pwndbg.aglib.regs.flags
            if flags is not None:
                for regname, flag in flags.items():
                    flag.update(regname)
                    regs.append(flag)
        if pwndbg.aglib.qemu.is_qemu_kernel() and pwndbg.aglib.regs.kernel is not None:
            if pwndbg.aglib.regs.kernel.segments is not None:
                regs.append(pwndbg.aglib.regs.kernel.segments)

    for reg in regs:
        if reg is None:
            continue
        # If it's a VisitableRegister which has special logic to determine what to print
        if not isinstance(reg, str):
            desc = reg.context(rc)
            if desc is not None:
                result.append(desc)
            continue

        # Resolve "sp" and "pc" to the real architectural register names
        reg = pwndbg.aglib.regs.current.resolve_aliases(reg)

        desc = rc.register_context_default(reg)
        if desc is not None:
            result.append(desc)

    return result


disasm_lines = pwndbg.config.add_param(
    "context-disasm-lines", 10, "number of additional lines to print in the disasm context"
)


def try_emulate_if_bug_disable(handler: Callable[[], T]) -> T:
    try:
        return handler()
    except U.UcError as e:
        print(
            message.warn(
                f"Warning: Emulation context disabled due to a Unicorn error: \n{str(e)}\n"
                "If you want to enable it again, use `set emulate on`."
            )
        )
        pwndbg.config.emulate.value = "off"
        return handler()


@serve_context_history
def context_disasm(
    target: OutputTarget = sys.stdout,
    with_banner: bool = True,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    flavor = pwndbg.dbg.x86_disassembly_flavor()
    syntax = pwndbg.aglib.disasm.disassembly.CapstoneSyntax[flavor]

    # Get the Capstone object to set disassembly syntax
    cs = next(
        iter(getattr(pwndbg.aglib.disasm.disassembly.get_disassembler, "cache").values()), None
    )

    # The `None` case happens when the cache was not filled yet (see e.g. #881)
    if cs is not None and cs.syntax != syntax:
        pwndbg.lib.cache.clear_caches()
        pwndbg.aglib.disasm.disassembly.computed_instruction_cache.clear()

    additional_disasm_lines = max(int(disasm_lines), height or 0)

    result = try_emulate_if_bug_disable(
        lambda: pwndbg.aglib.nearpc.nearpc(
            back_lines=additional_disasm_lines // 2,
            total_lines=additional_disasm_lines + 1,
            emulate=bool(not pwndbg.config.emulate == "off"),
            use_cache=True,
        )
    )

    # Note: we must fetch emulate value again after disasm since
    # we check if we can actually use emulation in `can_run_first_emulate`
    # and this call may disable it
    thumb_mode_str = get_thumb_mode_string()
    if thumb_mode_str is not None:
        info = f" / {pwndbg.aglib.arch.name} / {thumb_mode_str} mode / set emulate {pwndbg.config.emulate}"
    else:
        info = f" / {pwndbg.aglib.arch.name} / set emulate {pwndbg.config.emulate}"
    banner = [pwndbg.ui.banner("disasm", target=target, width=width, extra=info)]

    # If we didn't disassemble backward, try to make sure
    # that the amount of screen space taken is roughly constant.
    while len(result) < additional_disasm_lines + 1:
        result.append("")

    return banner + result if with_banner else result


theme.add_param("highlight-source", True, "whether to highlight the closest source line")
source_disasm_lines = pwndbg.config.add_param(
    "context-code-lines", 14, "number of source code lines to print by the context command"
)
pwndbg.config.add_param(
    "context-code-tabstop", 8, "number of spaces that a <tab> in the source code counts for"
)
theme.add_param("code-prefix", "►", "prefix marker for 'context code' command")
# All of these are also used for the decompilation context^^


def get_highlight_source_uncached(filename: str) -> tuple[str, ...]:
    with open(filename, encoding="utf-8", errors="ignore") as f:
        source = f.read()

    if pwndbg.config.syntax_highlight:
        source = H.syntax_highlight(source, filename)

    source_lines = source.split("\n")
    source_lines = tuple(line.rstrip() for line in source_lines)
    return source_lines


@pwndbg.lib.cache.cache_until("objfile")
def get_highlight_source(filename: str) -> tuple[str, ...]:
    # Notice that the code is cached
    return get_highlight_source_uncached(filename)


def get_filename_and_formatted_source(height: int | None = None) -> tuple[str, list[str], int]:
    """
    Returns formatted, lines limited and highlighted source as list
    or if it isn't there - an empty list
    """
    frame = pwndbg.dbg.selected_frame()
    if not frame:
        return "", [], 0

    sal = frame.sal()

    # Check if source code is available
    if sal is None:
        return "", [], 0

    # Get the full source code
    filename, closest_line = sal

    try:
        source = get_highlight_source(filename)
    except OSError:
        return "", [], closest_line

    if not source:
        return "", [], closest_line

    nlines = max(int(source_disasm_lines), height or 0)
    formatted_source = pretty_print.format_source(list(source), nlines, closest_line)

    return filename, formatted_source, closest_line


should_decompile = pwndbg.config.add_param(
    "context-integration-decompile",
    True,
    "whether context should fall back to decompilation with no source code",
)


@serve_context_history
def context_code(
    target: OutputTarget = sys.stdout,
    with_banner: bool = True,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    filename, formatted_source, line = get_filename_and_formatted_source(height)

    # Try getting source from files
    if formatted_source:
        bannerline = (
            [pwndbg.ui.banner("Source (code)", target=target, width=width)] if with_banner else []
        )
        return bannerline + [f"In file: {filename}:{line}"] + formatted_source

    if should_decompile and pwndbg.aglib.regs.pc is not None:
        nlines = max(int(source_disasm_lines), height or 0)
        # Will be None if we aren't connected or decompilation fails.
        code: list[str] | None = pwndbg.dintegration.manager.decompile_pretty(
            pwndbg.aglib.regs.pc, nlines
        )
        if code is None:
            return []

        bannerline = [pwndbg.ui.banner("Decomp", target=target, width=width)] if with_banner else []
        return bannerline + code
    return []


stack_lines = pwndbg.config.add_param(
    "context-stack-lines", 8, "number of lines to print in the stack context"
)


@serve_context_history
def context_stack(
    target: OutputTarget = sys.stdout,
    with_banner: bool = True,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    tui_stack_lines = max(int(stack_lines), height or 0)
    result = [pwndbg.ui.banner("stack", target=target, width=width)] if with_banner else []
    telescope = pwndbg.commands.telescope.telescope(
        pwndbg.aglib.regs.sp, to_string=True, count=tui_stack_lines
    )
    if telescope:
        result.extend(telescope)
    return result


backtrace_lines = pwndbg.config.add_param(
    "context-backtrace-lines", 8, "number of lines to print in the backtrace context"
)
backtrace_frame_label = theme.add_param(
    "backtrace-frame-label", "", "frame number label for backtrace"
)


@serve_context_history
def context_backtrace(
    with_banner: bool = True,
    target: OutputTarget = sys.stdout,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    result = []

    if with_banner:
        result.append(pwndbg.ui.banner("backtrace", target=target, width=width))

    this_frame = pwndbg.dbg.selected_frame()
    newest_frame = this_frame
    oldest_frame = this_frame

    tui_backtrace_lines = max(int(backtrace_lines), height or 0)
    for i in range(tui_backtrace_lines - 1):
        try:
            candidate = oldest_frame.parent()
        # We catch an error in case of a `gdb.error: PC not saved` case
        except pwndbg.dbg_mod.Error:
            break

        if not candidate:
            break
        oldest_frame = candidate

    for i in range(tui_backtrace_lines - 1):
        candidate = newest_frame.child()
        if not candidate:
            break
        newest_frame = candidate

    frame = newest_frame
    i = 0
    bt_prefix = f"{pwndbg.config.backtrace_prefix}"
    while True:
        prefix = bt_prefix if frame == this_frame else " " * len(bt_prefix)
        prefix = f" {c.prefix(prefix)}"
        addrsz = c.address(pwndbg.ui.addrsz(frame.pc()))
        symbol = c.symbol(pwndbg.aglib.symbol.resolve_addr(int(frame.pc())))
        if symbol:
            if bool(config_backtrace_format):
                offset_regex = re.compile(r"^(.+)\+(\d+)$")
                parts = offset_regex.match(symbol)
                if parts:
                    symbol = f"{parts[1]}+{int(parts[2]):#x}"
            addrsz = f"{addrsz} {symbol}"
        result.append(f"{prefix} {c.frame_label(f'{backtrace_frame_label}{i}')} {addrsz}")

        if frame == oldest_frame:
            break

        frame = frame.parent()
        i += 1
    return result


@serve_context_history
def context_args(
    with_banner: bool = True,
    target: OutputTarget = sys.stdout,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    args = pwndbg.arguments.format_args(pwndbg.aglib.disasm.disassembly.one())

    # early exit to skip section if no arg found
    if not args:
        return []

    if with_banner:
        args.insert(0, pwndbg.ui.banner("arguments", target=target, width=width))

    return args


last_signal: list[str] = []

thread_status_messages = {
    "running": pwndbg.color.light_green("running"),
    "stopped": pwndbg.color.yellow("stopped"),
    "exited": pwndbg.color.gray("exited "),
}


def get_thread_status(thread: gdb.InferiorThread) -> str:
    if thread.is_running():
        return thread_status_messages["running"]
    if thread.is_stopped():
        return thread_status_messages["stopped"]
    if thread.is_exited():
        return thread_status_messages["exited"]
    return "unknown"


@serve_context_history
def context_threads(
    with_banner: bool = True,
    target: OutputTarget = sys.stdout,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    try:
        original_thread = gdb.selected_thread()
    except SystemError:
        original_thread = None
    try:
        original_frame = gdb.selected_frame()
    except gdb.error:
        original_frame = None

    all_threads = gdb.selected_inferior().threads()[::-1]

    displayed_threads = []

    if original_thread is not None and original_thread.is_valid():
        displayed_threads.append(original_thread)

    max_threads_display = max(int(config_max_threads_display), height or 0)
    for thread in all_threads:
        if len(displayed_threads) >= max_threads_display:
            break

        if thread.is_valid() and thread is not original_thread:
            displayed_threads.append(thread)

    num_threads_not_shown = len(all_threads) - len(displayed_threads)

    if len(displayed_threads) < 2:
        return []

    out = (
        [pwndbg.ui.banner(f"threads ({len(all_threads)} total)", target=target, width=width)]
        if with_banner
        else []
    )
    max_name_length = 0
    max_global_num_len = 0

    for thread in displayed_threads:
        name = thread.name or ""
        if len(name) > max_name_length:
            max_name_length = len(name)
        if len(str(thread.global_num)) > max_global_num_len:
            max_global_num_len = len(str(thread.global_num))

    for thread in filter(lambda t: t.is_valid(), displayed_threads):
        selected = " ►" if thread is original_thread else "  "
        name = thread.name if thread.name is not None else ""
        name_padding = max_name_length - len(name)
        global_num_padding = max(2, max_global_num_len - len(str(thread.global_num)))
        status = get_thread_status(thread)

        line = (
            f" {selected} {thread.global_num} "
            f"{' ' * global_num_padding}"
            f'"{pwndbg.color.cyan(name)}" '
            f"{' ' * name_padding}"
            f"{status}: "
        )

        if thread.is_stopped():
            thread.switch()
            pc = gdb.selected_frame().pc()

            pc_colored = mem_color.get(pc)
            symbol = pwndbg.aglib.symbol.resolve_addr(int(pc))

            line += f"{pc_colored}"
            if symbol:
                line += f" <{pwndbg.color.bold(pwndbg.color.green(symbol))}> "

        out.append(line)

    if num_threads_not_shown:
        out.append(
            pwndbg.lib.tips.color_tip(
                f"Not showing {num_threads_not_shown} thread(s). Use `set context-max-threads <number of threads>` to change this."
            )
        )

    if original_thread is not None and original_thread.is_valid():
        original_thread.switch()
    if original_frame is not None and original_frame.is_valid():
        original_frame.select()

    return out


@pwndbg.dbg.event_handler(EventType.STOP, EventHandlerPriority.SAVE_SIGNAL)
@pwndbg.dbg.event_handler(EventType.EXIT, EventHandlerPriority.SAVE_SIGNAL)
def save_signal() -> None:
    global last_signal
    last_signal = result = []

    if pwndbg.dbg.is_gdblib_available() and _is_rr_present():
        # When users use rr (https://rr-project.org or https://github.com/mozilla/rr)
        # we can't access $_siginfo, so lets just show current pc
        # see also issue 476
        if pwndbg.aglib.regs.pc is not None:
            result.append(message.signal(f"current pc: {pwndbg.aglib.regs.pc:#x}"))
        return

    process = pwndbg.dbg.selected_inferior()
    if not process:
        return

    if not (process.stopped_with_signal() or process.stopped_at_breakpoint()):
        return

    signal = pwndbg.aglib.signal.get_last_signal()
    if signal is None:
        return
    msg = f"Program received signal {signal}"

    if signal == "SIGSEGV":
        try:
            desc_short, desc_long = pwndbg.aglib.signal.get_segv_information()
            msg = f"Program received signal {desc_short}"
            if desc_long:
                msg += desc_long
        except pwndbg.dbg_mod.Error:
            pass
    elif signal == "SIGTRAP":
        result.append(message.breakpoint(f"Breakpoint hit at {pwndbg.aglib.regs.pc:#x}"))
        return
    result.append(message.signal(msg))


@serve_context_history
def context_last_signal(
    with_banner: bool = True,
    target: OutputTarget = sys.stdout,
    width: int | None = None,
    height: int | None = None,
) -> list[str]:
    if not last_signal:
        return []

    result = last_signal[::]
    if with_banner:
        result.insert(0, pwndbg.ui.banner("last signal", target=target, width=width))

    return result


context_sections: dict[str, Callable[..., list[str]]] = {
    "a": context_args,
    "r": context_regs,
    "d": context_disasm,
    "s": context_stack,
    "b": context_backtrace,
    "c": context_code,
    "l": context_last_signal,
}


if pwndbg.dbg.is_gdblib_available():
    # Add the sections that need gdblib.
    context_sections = {
        **context_sections,
        "e": context_expressions,
        "h": context_heap_tracker,
        "t": context_threads,
    }


@pwndbg.lib.cache.cache_until("forever")
def _is_rr_present() -> bool:
    """
    Checks whether rr project is present (so someone launched e.g. `rr replay <some-recording>`)
    """

    # this is ugly but I couldn't find a better way to do it
    # feel free to refactor it
    globals_list_literal_str = gdb.execute("python print(list(globals().keys()))", to_string=True)
    interpreter_globals = ast.literal_eval(globals_list_literal_str)

    return "RRCmd" in interpreter_globals and "RRWhere" in interpreter_globals



from __future__ import annotations

from types import TracebackType

import pwndbg.color as color


class IndentContextManager:
    def __init__(self) -> None:
        self.indent = 0

    def __enter__(self) -> None:
        self.indent += 1

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.indent -= 1
        assert self.indent >= 0

    def print(self, *a, **kw) -> None:
        print("    " * self.indent, *a, **kw)

    def addr_hex(self, val: int) -> str:
        return color.yellow(hex(val))

    def aux_hex(self, val: int) -> str:
        return color.red(hex(val))

    def prefix(self, s: str):
        if self.indent % 2 == 0:
            return color.blue(s)
        return color.green(s)



from __future__ import annotations

import argparse

import pwndbg
import pwndbg.commands
from pwndbg.commands import CommandCategory
from pwndbg.lib.tips import color_tip
from pwndbg.lib.tips import get_all_tips
from pwndbg.lib.tips import get_tip_of_the_day

parser = argparse.ArgumentParser(description="Shows tips.")
parser.add_argument("-a", "--all", action="store_true", help="Show all tips.")


@pwndbg.commands.Command(parser, category=CommandCategory.MISC)
def tips(all: bool) -> None:
    if all:
        for tip in get_all_tips(pwndbg.dbg.name().value):
            print(color_tip(tip))
    else:
        print(color_tip(get_tip_of_the_day(pwndbg.dbg.name().value)))



"""
Commands for dealing with Linux kernel slab allocator. Currently, only SLUB is supported.

Some of the code here was inspired from https://github.com/NeatMonster/slabdbg
Some of the code here was inspired from https://github.com/osandov/drgn
"""

from __future__ import annotations

import argparse

from tabulate import tabulate

import pwndbg
import pwndbg.aglib.kernel.slab
import pwndbg.aglib.memory
import pwndbg.color
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.aglib.kernel.slab import CpuCache
from pwndbg.aglib.kernel.slab import Freelist
from pwndbg.aglib.kernel.slab import NodeCache
from pwndbg.aglib.kernel.slab import Slab
from pwndbg.aglib.kernel.slab import find_containing_slab_cache
from pwndbg.commands import CommandCategory
from pwndbg.lib.exception import IndentContextManager

parser = argparse.ArgumentParser(
    description="Prints information about the linux kernel's slab allocator SLUB."
)
subparsers = parser.add_subparsers(dest="command")
subparsers.required = True


parser_list = subparsers.add_parser("list", description="List SLUB caches filtered by name.")
parser_list.add_argument(
    "filter_",
    metavar="filter",
    type=str,
    default=None,
    nargs="?",
    help="Only show caches that contain the given filter string",
)

parser_info = subparsers.add_parser("info", description="Dump information about a cache.")
parser_info.add_argument("names", metavar="name", type=str, nargs="+", help="")
parser_info.add_argument("-v", "--verbose", action="store_true", help="")
parser_info.add_argument("-c", "--cpu", type=int, help="CPU to display")
parser_info.add_argument("-n", "--node", type=int, help="")
parser_info.add_argument("-p", "--partial", action="store_true", help="displays partial lists")
parser_info.add_argument("-a", "--active", action="store_true", help="displays the active list")

parser_contains = subparsers.add_parser("contains", description="Get the cache for an address.")
parser_contains.add_argument("addresses", metavar="addr", type=str, nargs="+", help="")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelSymbols
@pwndbg.commands.OnlyWhenPagingEnabled
def slab(
    command: str,
    filter_: str | None = None,
    names: list[str] | None = None,
    verbose: bool = False,
    addresses: list[str] | None = None,
    cpu: int | None = None,
    node: int | None = None,
    partial: bool = False,
    active: bool = False,
) -> None:
    pwndbg.aglib.kernel.slab.recover_slab_typeinfo()
    if command == "list":
        slab_list(filter_)
    elif command == "info":
        assert names
        if not partial and not active:
            partial = active = True
        for name in names:
            slab_info(name, verbose, cpu, node, active, partial)
    elif command == "contains":
        assert addresses
        for addr in addresses:
            slab_contains(addr)


def emphasize(s):
    return pwndbg.color.underline(pwndbg.color.bold(pwndbg.color.red(s)))


indent = IndentContextManager()


def handle_next(curr: int, freelist: Freelist) -> str:
    next = freelist.find_next(curr)
    if next == 0:
        return "no next"
    desc = f"next: {indent.aux_hex(next)}"
    if not pwndbg.aglib.memory.is_kernel(next + freelist.offset):
        desc = emphasize("invalid address") + " " + desc
    elif freelist.cyclic is not None and freelist.cyclic == curr:
        desc = emphasize("cyclic list detected") + ", " + desc
    elif not freelist.slab or next not in freelist.slab:
        desc = emphasize("next is not within the slab") + ", " + desc
    elif not freelist.is_valid_obj(next):
        desc = emphasize("unaligned or out-of-range") + " " + desc
    return desc


def freelist_desc(freelist: Freelist) -> str:
    head = int(freelist)
    desc = None
    if head:
        if not pwndbg.aglib.memory.is_kernel(head):
            desc = "invalid address"
        elif not freelist.slab or head not in freelist.slab:
            desc = "not within the slab"
        elif not freelist.is_valid_obj(head):
            desc = "unaligned or out-of-range"
    return indent.addr_hex(head) + (f" [{emphasize(desc)}]" if desc else "")


def print_slab(slab: Slab, verbose: bool) -> None:
    indent.print(
        f"- {indent.prefix('Slab')} @ {indent.addr_hex(slab.virt_address)} [{indent.aux_hex(slab.slab_address)}]:"
    )

    with indent:
        indent.print(f"{indent.prefix('In-Use')}: {slab.inuse}/{slab.object_count}")
        indent.print(f"{indent.prefix('Frozen')}: {slab.frozen}")
        indent.print(f"{indent.prefix('Freelist')}: {freelist_desc(slab.freelist)}")

        cpu_freelist = slab.cpu_cache.freelist if slab.is_active else None
        indexes = {}
        freelist = slab.freelist
        for idx, addr in enumerate(freelist):
            if addr in indexes:
                break
            indexes[addr] = idx
            idx += 1
        if cpu_freelist is not None:
            for idx, addr in enumerate(cpu_freelist):
                if addr in indexes:
                    break
                indexes[addr] = idx

        if verbose:
            with indent:
                free_objects = slab.free_objects
                for addr in slab.objects:
                    prefix = f"- {indent.prefix('[0x--]')} {hex(addr)}"
                    if addr not in free_objects:
                        indent.print(f"{prefix} (in-use)")
                        continue
                    index = indexes[addr]
                    if addr in indexes:
                        prefix = f"- {indent.prefix(f'[0x{index:02x}]')} {indent.addr_hex(addr)}"
                    desc = None
                    in_cpu_freelist = False
                    if addr in freelist:
                        desc = handle_next(addr, freelist)
                    elif cpu_freelist is not None and addr in cpu_freelist:
                        # need to traverse the list to catch potential freelist.cyclic
                        desc = handle_next(addr, cpu_freelist)
                        in_cpu_freelist = True
                    if desc is None:
                        desc = "something went wrong"
                    if in_cpu_freelist:
                        indent.print(f"{prefix} ({desc}) [CPU cache]")
                        continue
                    indent.print(f"{prefix} ({desc})")


def print_cpu_cache(cpu_cache: CpuCache, verbose: bool, active: bool, partial: bool) -> None:
    indent.print(
        f"{indent.prefix('kmem_cache_cpu')} @ {indent.addr_hex(cpu_cache.address)} [CPU {cpu_cache.cpu}]:"
    )
    with indent:
        if active:
            indent.print(f"{indent.prefix('Freelist')}:", freelist_desc(cpu_cache.freelist))
            active_slab = cpu_cache.active_slab
            if active_slab:
                indent.print(f"{indent.prefix('Active Slab')}:")
                with indent:
                    print_slab(active_slab, verbose)
            else:
                indent.print("Active Slab: (none)")

        if not partial:
            return
        partial_slabs = cpu_cache.partial_slabs
        if not partial_slabs:
            indent.print("Partial Slabs: (none)")
            return
        slabs = partial_slabs[0].slabs
        # the kernel checks cpu_partial_slabs to determine whether partial slabs are to be flushed
        # see: https://elixir.bootlin.com/linux/v6.13/source/mm/slub.c#L3209
        cpu_partial_slabs = partial_slabs[0].slab_cache.cpu_partial_slabs
        if cpu_partial_slabs is None:
            # legacy
            cpu_partial_slabs = partial_slabs[0].pobjects
        indent.print(
            f"{indent.prefix('Partial Slabs')} [nr_slabs/cpu_partial_slabs: {indent.aux_hex(slabs)}/{indent.aux_hex(cpu_partial_slabs)}]"
        )
        with indent:
            for partial_slab in partial_slabs:
                print_slab(partial_slab, verbose)


def print_node_cache(node_cache: NodeCache, verbose: bool) -> None:
    address, nr_partial, min_partial, node = (
        node_cache.address,
        node_cache.nr_partial,
        node_cache.min_partial,
        node_cache.node,
    )
    # https://elixir.bootlin.com/linux/v6.13/source/mm/slub.c#L3140
    indent.print(
        f"{indent.prefix('kmem_cache_node')} @ {indent.addr_hex(address)} [NUMA node {node}, nr_partial/min_partial: {indent.aux_hex(nr_partial)}/{indent.aux_hex(min_partial)}]:"
    )
    with indent:
        partial_slabs = node_cache.partial_slabs
        if not partial_slabs:
            indent.print("Partial Slabs: (none)")
            return

        indent.print(
            f"{indent.prefix('Partial Slabs')} [nr_partial: {indent.aux_hex(len(partial_slabs))}]"
        )
        with indent:
            for slab in partial_slabs:
                print_slab(slab, verbose)


def slab_info(
    name: str, verbose: bool, cpu: int | None, node: int | None, active: bool, partial: bool
) -> None:
    slab_cache = pwndbg.aglib.kernel.slab.get_cache(name)

    if slab_cache is None:
        print(message.error(f"Cache {name} not found"))
        return

    indent.print(f"{indent.prefix('Slab Cache')} @ {indent.addr_hex(slab_cache.address)}")
    with indent:
        indent.print(f"{indent.prefix('Name')}: {slab_cache.name}")
        flags_list = slab_cache.flags
        if flags_list:
            indent.print(f"{indent.prefix('Flags')}: {' | '.join(flags_list)}")
        else:
            indent.print(f"{indent.prefix('Flags')}: (none)")

        indent.print(f"{indent.prefix('Offset')}: {indent.aux_hex(slab_cache.offset)}")
        indent.print(f"{indent.prefix('Slab size')}: {indent.aux_hex(slab_cache.slab_size)}")
        indent.print(
            f"{indent.prefix('Size (including metadata)')}: {indent.aux_hex(slab_cache.size)}"
        )
        indent.print(f"{indent.prefix('Align')}: {indent.aux_hex(slab_cache.align)}")
        indent.print(f"{indent.prefix('Object Size')}: {indent.aux_hex(slab_cache.object_size)}")
        useroffset, usersize = slab_cache.useroffset, slab_cache.usersize
        if useroffset is not None and usersize is not None:
            indent.print(f"{indent.prefix('Usercopy region offset')}: {useroffset}")
            indent.print(f"{indent.prefix('Usercopy region size')}: {usersize}")

        for cpu_cache in slab_cache.cpu_caches:
            if cpu is not None and cpu_cache.cpu != cpu:
                continue
            print_cpu_cache(cpu_cache, verbose, active, partial)

        if not partial:
            return

        for node_cache in slab_cache.node_caches:
            if node is not None and node != node_cache.node:
                continue
            print_node_cache(node_cache, verbose)


def slab_list(filter_) -> None:
    results = [
        [
            slab_cache.name,
            slab_cache.oo_objects,
            slab_cache.size,
            slab_cache.object_size,
            slab_cache.inuse,
            slab_cache.oo_order,
        ]
        for slab_cache in pwndbg.aglib.kernel.slab.caches()
        if not filter_ or filter_ in slab_cache.name
    ]

    print(tabulate(results, headers=["Name", "# Objects", "Size", "Obj Size", "# inuse", "order"]))


def slab_contains(address: str) -> None:
    """prints the slab_cache associated with the provided address"""

    addr = None
    try:
        addr = int(pwndbg.dbg.selected_frame().evaluate_expression(address))
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"Could not parse '{address}'"))
        print(message.error(f"Message: {e}"))
        return

    try:
        base, slab_cache = find_containing_slab_cache(addr)
        assert base and slab_cache, "cannot find the kmem_cache the address belongs to."
        addr = base + ((addr - base) // slab_cache.size) * slab_cache.size
        indent.print(f"{addr:#x} @", message.hint(f"{slab_cache.name}"))
        desc = "[inactive]"
        inuse = f"[something went wrong: {hex(addr)}]"
        slab = slab_cache.find_containing_slab(addr)
        if slab:
            if addr in slab.free_objects:
                inuse = "free"
            elif addr in slab.objects:
                inuse = "in-use"
            if slab.is_active:
                if slab.is_cpu:
                    desc = f"[active, cpu {slab.cpu_cache.cpu}]"
            else:
                if slab.is_cpu:
                    desc = f"[partial, cpu {slab.cpu_cache.cpu}]"
                else:
                    desc = f"[partial, node {slab.node_cache.node}]"
        else:
            inuse = "in-use"
        indent.print("slab:", message.hint(f"{hex(base)}"), desc)
        indent.print("status:", message.hint(inuse))
    except Exception as e:
        print(message.warn(f"address does not belong to a SLUB cache: {e}"))
